from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue
from ...event import RoundEndEventArguments

from ...action import Actions, ChangeObjectUsageAction, MakeDamageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, 
    ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class YueguiThrowingMode_4_1(AttackerSummonBase):
    name: Literal['Yuegui: Throwing Mode'] = 'Yuegui: Throwing Mode'
    version: Literal['4.1'] = '4.1'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    damage: int = 1
    is_talent_activated: bool = False

    def renew(self, obj: 'YueguiThrowingMode_4_1') -> None:
        self.is_talent_activated = obj.is_talent_activated
        super().renew(obj)

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        """
        When in round end, attack enemy active charactor and heal our 
        charactor. When talent activated, and usage is 1, damage is increased
        to 2.
        """
        if self.usage == 1 and self.is_talent_activated:
            self.damage = 2
        ret = super().event_handler_ROUND_END(event, match)
        table = match.player_tables[self.position.player_idx]
        charactors = table.charactors
        # select charactor with most damage taken
        selected_charactor = charactors[table.active_charactor_idx]
        for c in charactors:
            if c.is_alive and c.damage_taken > selected_charactor.damage_taken:
                selected_charactor = c
        # heal charactor, and do default actions
        return [MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = selected_charactor.position,
                    damage = -self.damage,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost(),
                )
            ],
        )] + ret


# Skills


class RaphanusSkyCluster(ElementalSkillBase):
    name: Literal['Raphanus Sky Cluster'] = 'Raphanus Sky Cluster'
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        create object
        """
        args = {}
        if self.is_talent_equipped(match):
            args['is_talent_activated'] = True
        return [
            self.create_summon('Yuegui: Throwing Mode', args),
            self.charge_self(1)
        ]


class MoonjadeDescent(ElementalBurstBase):
    name: Literal['Moonjade Descent'] = 'Moonjade Descent'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_team_status('Adeptal Legacy')
        ]


# Talents


class Beneficent_4_1(SkillTalent):
    name: Literal['Beneficent']
    version: Literal['4.1'] = '4.1'
    charactor_name: Literal['Yaoyao'] = 'Yaoyao'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3,
    )
    skill: Literal['Raphanus Sky Cluster'] = 'Raphanus Sky Cluster'


# charactor base


class Yaoyao_4_1(CharactorBase):
    name: Literal['Yaoyao']
    version: Literal['4.1'] = '4.1'
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | RaphanusSkyCluster | MoonjadeDescent
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = "Toss 'N' Turn Spear",
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            RaphanusSkyCluster(),
            MoonjadeDescent(),
        ]


register_class(Yaoyao_4_1 | Beneficent_4_1 | YueguiThrowingMode_4_1)



from typing import List, Literal

from ....utils.class_registry import register_class

from ..charactor_base import PhysicalNormalAttackBase

from ...consts import DieColor

from ...struct import Cost
from .jean_3_3 import (
    Jean_3_3, DandelionField_3_3, LandsOfDandelion_3_3, GaleBlade,
    DandelionBreeze as DB_3_3
)


class DandelionField_4_2(DandelionField_3_3):
    version: Literal['4.2'] = '4.2'
    damage: int = 1


class LandsOfDandelion_4_2(LandsOfDandelion_3_3):
    version: Literal['4.2'] = '4.2'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 4,
        charge = 2
    )


class DandelionBreeze(DB_3_3):
    version: Literal['4.2'] = '4.2'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 4,
        charge = 2
    )


class Jean_4_2(Jean_3_3):
    version: Literal['4.2'] = '4.2'
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | GaleBlade | DandelionBreeze
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Favonius Bladework',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            GaleBlade(),
            DandelionBreeze(),
        ]


register_class(Jean_4_2 | DandelionField_4_2 | LandsOfDandelion_4_2)



from typing import List, Literal

from ....utils.class_registry import register_class

from ..charactor_base import SkillTalent
from ...consts import DieColor
from ...struct import Cost
from .yoimiya_3_8 import Yoimiya_3_8 as Y_3_8
from .yoimiya_3_8 import RyuukinSaxifrage as RS_3_8
from .yoimiya_3_8 import FireworkFlareUp, NiwabiFireDance


class RyuukinSaxifrage(RS_3_8):
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3,
        charge = 2
    )


class NaganoharaMeteorSwarm_3_3(SkillTalent):
    name: Literal['Naganohara Meteor Swarm']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Yoimiya'] = 'Yoimiya'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 2,
    )
    skill: Literal['Niwabi Fire-Dance'] = 'Niwabi Fire-Dance'
    status_max_usage: int = 2


class Yoimiya_3_3(Y_3_8):
    version: Literal['3.3']
    max_charge: int = 2
    skills: List[FireworkFlareUp | NiwabiFireDance | RyuukinSaxifrage] = []

    def _init_skills(self) -> None:
        self.skills = [
            FireworkFlareUp(),
            NiwabiFireDance(),
            RyuukinSaxifrage()
        ]


register_class(Yoimiya_3_3 | NaganoharaMeteorSwarm_3_3)




from typing import Any, Literal

from ....utils.class_registry import register_class

from ...consts import DamageElementalType, IconType, SkillType

from ...modifiable_values import DamageIncreaseValue
from .base import (
    ElementalInfusionCharactorStatus, RoundCharactorStatus, 
    UsageCharactorStatus
)


class Grimheart_3_8(UsageCharactorStatus):
    name: Literal['Grimheart'] = 'Grimheart'
    version: Literal['3.8'] = '3.8'
    damage: int = 3
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.ATK_UP_ICE] = IconType.ATK_UP_ICE

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if self.version == '3.5':
            self.damage = 2
        else:
            assert self.version == '3.8'
            self.damage = 3

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        """
        When character uses Icetide Vortex, increase damage by self.damage
        """
        if value.damage_from_element_reaction:
            # from element reaction, do nothing
            return value
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_SKILL
        ):
            # not self use elemental skill, do nothing
            return value
        assert self.usage > 0
        assert mode == 'REAL'
        self.usage -= 1
        value.damage += self.damage
        return value


class Grimheart_3_5(Grimheart_3_8):
    version: Literal['3.5']
    damage: int = 2


class CryoFusionAyaka_3_3(ElementalInfusionCharactorStatus,
                          RoundCharactorStatus):
    """
    Inherit from vanilla elemental infusion status, and has talent activated
    argument. when talent activated, it will gain electro damage +1 buff,
    and usage +1.
    """
    name: Literal['Cryo Elemental Infusion'] = 'Cryo Elemental Infusion'
    mark: Literal['Kamisato Ayaka']  # used to select right status
    desc: Literal['', 'ayaka_talent'] = ''
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1

    talent_activated: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.talent_activated:
            self.desc = 'ayaka_talent'

    def renew(self, new_status: 'CryoFusionAyaka_3_3') -> None:
        super().renew(new_status)
        if new_status.talent_activated:
            self.talent_activated = True
            self.desc = 'ayaka_talent'

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        If talent activated, increase cryo damage dealt by this charactor 
        by 1.
        """
        if not self.talent_activated:
            # talent not activated, do nothing
            return value
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            # not corresponding charactor use skill, do nothing
            return value
        if (
            value.damage_elemental_type != DamageElementalType.CRYO
        ):  # pragma: no cover
            # not cryo damage, do nothing
            return value
        assert mode == 'REAL'
        value.damage += 1
        return value


register_class(Grimheart_3_8 | Grimheart_3_5 | CryoFusionAyaka_3_3)



from typing import Any, Literal

from ....utils.class_registry import register_class

from ...consts import IconType

from ...modifiable_values import DamageIncreaseValue
from .base import RoundTeamStatus, ShieldTeamStatus


class RebelliousShield_3_7(ShieldTeamStatus):
    name: Literal['Rebellious Shield'] = 'Rebellious Shield'
    version: Literal['3.7'] = '3.7'
    usage: int = 1
    max_usage: int = 2


class MillennialMovementFarewellSong_3_7(RoundTeamStatus):
    name: Literal[
        'Millennial Movement: Farewell Song'
    ] = 'Millennial Movement: Farewell Song'
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        if value.is_corresponding_charactor_use_damage_skill(
            self.position, match, None
        ):
            assert mode == 'REAL'
            value.damage += 1
        return value


register_class(RebelliousShield_3_7 | MillennialMovementFarewellSong_3_7)



from typing import Dict, List, Literal

from ....event import (
    ReceiveDamageEventArguments, RoundPrepareEventArguments, 
    UseSkillEventArguments
)
from .....utils.class_registry import register_class
from ....status.charactor_status.base import DefendCharactorStatus
from ....summon.base import AttackAndGenerateStatusSummonBase
from ....action import Actions, ChargeAction, CreateObjectAction
from ....match import Match
from ....struct import Cost
from ....consts import (
    ELEMENT_TO_DAMAGE_TYPE, DamageElementalType, DieColor, ElementType, 
    FactionType, ObjectPositionType, SkillType, WeaponType
)
from ....charactor.charactor_base import (
    CharactorBase, ElementalBurstBase, ElementalNormalAttackBase, 
    ElementalSkillBase, PassiveSkillBase, SkillBase, SkillTalent
)
from .....utils.desc_registry import DescDictType


class PyroScorpionGuardianStance_4_3(DefendCharactorStatus):
    name: Literal[
        'Pyro Scorpion: Guardian Stance'
    ] = 'Pyro Scorpion: Guardian Stance'
    version: Literal['4.3'] = '4.3'
    desc: Literal['', 'talent'] = ''

    # usage and max usage will be set by summon
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1


class SpiritOfOmenPyroScorpion_4_3(AttackAndGenerateStatusSummonBase):
    name: Literal[
        'Spirit of Omen: Pyro Scorpion'
    ] = 'Spirit of Omen: Pyro Scorpion'
    desc: Literal['', 'talent'] = ''
    version: Literal['4.3'] = '4.3'
    usage: int = 2
    max_usage: int = 2
    damage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    status_name: str = 'Pyro Scorpion: Guardian Stance'

    def renew(self, obj: 'SpiritOfOmenPyroScorpion_4_3') -> None:
        super().renew(obj)
        if obj.desc == 'talent' and self.desc != 'talent':
            self.desc = 'talent'

    def _create_status(self, match: Match) -> List[CreateObjectAction]:
        """
        Create for eremite self
        """
        charactors = match.player_tables[self.position.player_idx].charactors
        for c in charactors:
            if c.name == 'Eremite Scorching Loremaster':
                usage = 1
                if self.desc == 'talent':
                    usage = 2
                return [
                    CreateObjectAction(
                        object_name = 'Pyro Scorpion: Guardian Stance',
                        object_position = c.position.set_area(
                            ObjectPositionType.CHARACTOR_STATUS),
                        object_arguments = {
                            'desc': self.desc,
                            'usage': usage,
                            'max_usage': usage,
                        }
                    )
                ]
        else:
            raise AssertionError('Eremite Scorching Loremaster not found')

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[CreateObjectAction]:
        """
        Reset damage and call super
        """
        self.damage = 1
        return super().event_handler_ROUND_PREPARE(event, match)

    def event_handler_USE_SKILL(
        self, event: UseSkillEventArguments, match: Match
    ) -> List[Actions]:
        """
        If talent activated, and self eremite use normal or elemental skill,
        damage be 2.
        """
        if self.desc != 'talent':
            # not talent, do nothing
            return []
        skill_position = event.action.skill_position
        if skill_position.player_idx != self.position.player_idx:
            # not self charactor use skill, do nothing
            return []
        charactor = match.player_tables[skill_position.player_idx].charactors[
            skill_position.charactor_idx]
        if charactor.name != 'Eremite Scorching Loremaster':
            # not eremite use skill, do nothing
            return []
        skill: SkillBase = match.get_object(skill_position)  # type: ignore
        assert skill is not None
        if (
            skill.skill_type == SkillType.NORMAL_ATTACK
            or skill.skill_type == SkillType.ELEMENTAL_SKILL
        ):
            # is self eremite use normal or elemental skill, damage be 2
            self.damage = 2
        return []


class SpiritOfOmensAwakeningPyroScorpion(ElementalBurstBase):
    name: Literal[
        'Spirit of Omen\'s Awakening: Pyro Scorpion'
    ] = 'Spirit of Omen\'s Awakening: Pyro Scorpion'
    damage: int = 2
    damage_type: Literal[DamageElementalType.PYRO] = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        summon_desc = ''
        if self.is_talent_equipped(match):
            summon_desc = 'talent'
        return super().get_actions(match) + [
            self.create_summon('Spirit of Omen: Pyro Scorpion', {
                'desc': summon_desc
            })
        ]


class SpiritOfOmensPower(PassiveSkillBase):
    name: Literal['Spirit of Omen\'s Power'] = 'Spirit of Omen\'s Power'
    triggered: bool = False

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[ChargeAction]:
        """
        When self receive damage, and self alive, and self hp <= 7,
        and not triggered, gain 1 charge
        """
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        if not event.final_damage.is_corresponding_charactor_receive_damage(
            charactor.position, match, ignore_piercing = False
        ):
            # not self receive damage, do nothing
            return []
        if charactor.is_defeated:  # pragma: no cover
            # self defeated, do nothing
            return []
        if charactor.hp > 7:
            # self hp > 7, do nothing
            return []
        if self.triggered:
            # already triggered, do nothing
            return []
        self.triggered = True
        return [ ChargeAction(
            player_idx = charactor.position.player_idx,
            charactor_idx = charactor.position.charactor_idx,
            charge = 1
        )]


class Scorpocalypse_4_3(SkillTalent):
    name: Literal['Scorpocalypse'] = 'Scorpocalypse'
    version: Literal['4.3'] = '4.3'
    charactor_name: Literal[
        'Eremite Scorching Loremaster'
    ] = 'Eremite Scorching Loremaster'
    skill: Literal[
        'Spirit of Omen\'s Awakening: Pyro Scorpion'
    ] = 'Spirit of Omen\'s Awakening: Pyro Scorpion'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3,
        charge = 2
    )


class EremiteScorchingLoremaster_4_3(CharactorBase):
    name: Literal[
        'Eremite Scorching Loremaster'
    ] = 'Eremite Scorching Loremaster'
    version: Literal['4.3'] = '4.3'
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | ElementalSkillBase 
        | SpiritOfOmensAwakeningPyroScorpion | SpiritOfOmensPower
    ] = []
    faction: List[FactionType] = [FactionType.THE_EREMITES]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Searing Glare',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name = 'Blazing Strike',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalSkillBase.get_cost(self.element),
            ),
            SpiritOfOmensAwakeningPyroScorpion(),
            SpiritOfOmensPower(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTOR/Eremite Scorching Loremaster": {
        "names": {
            "en-US": "Eremite Scorching Loremaster",
            "zh-CN": "镀金旅团·炽沙叙事人"
        },
        "image_path": "cardface/Char_Monster_Muscleman.png",  # noqa: E501
        "id": 2303,
        "descs": {
            "4.3": {
                "en-US": "",
                "zh-CN": ""
            }
        }
    },
    "SKILL_Eremite Scorching Loremaster_NORMAL_ATTACK/Searing Glare": {
        "names": {
            "en-US": "Searing Glare",
            "zh-CN": "烧蚀之光"
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 1 Pyro DMG.",
                "zh-CN": "造成1点火元素伤害。"
            }
        }
    },
    "SKILL_Eremite Scorching Loremaster_ELEMENTAL_SKILL/Blazing Strike": {
        "names": {
            "en-US": "Blazing Strike",
            "zh-CN": "炎晶迸击"
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 3 Pyro DMG.",
                "zh-CN": "造成3点火元素伤害。"
            }
        }
    },
    "SKILL_Eremite Scorching Loremaster_ELEMENTAL_BURST/Spirit of Omen's Awakening: Pyro Scorpion": {  # noqa: E501
        "names": {
            "en-US": "Spirit of Omen's Awakening: Pyro Scorpion",
            "zh-CN": "厄灵苏醒·炎之魔蝎"
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Pyro DMG, summons 1 Spirit of Omen: Pyro Scorpion.",  # noqa: E501
                "zh-CN": "造成2点火元素伤害，召唤厄灵·炎之魔蝎。"
            }
        }
    },
    "SUMMON/Spirit of Omen: Pyro Scorpion": {
        "names": {
            "en-US": "Spirit of Omen: Pyro Scorpion",
            "zh-CN": "厄灵·炎之魔蝎"
        },
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deal 1 Pyro DMG.\nUsage(s): 2\n\nWhen entering play, when the Action Phase starts: Equips Pyro Scorpion: Guardian Stance to your Eremite Scorching Loremaster. (When Spirit of Omen: Pyro Scorpion is in play, the character will take 1 less DMG, once per Round.)",  # noqa: E501
                "zh-CN": "结束阶段：造成1点火元素伤害。\n可用次数：2\n\n入场时和行动阶段开始：使我方镀金旅团·炽沙叙事人附属炎之魔蝎·守势。（厄灵·炎之魔蝎在场时每回合1次，使角色受到的伤害-1。）"  # noqa: E501
            }
        }
    },
    "SUMMON/Spirit of Omen: Pyro Scorpion_talent": {
        "names": {
            "en-US": "Spirit of Omen: Pyro Scorpion",
            "zh-CN": "厄灵·炎之魔蝎"
        },
        "descs": {
            "4.3": {
                "en-US": "End Phase: Deal 1 Pyro DMG; in Rounds when Eremite Scorching Loremaster has used a Normal Attack or Elemental Skill, the damage dealt +1.\nUsage(s): 2\n\nWhen entering play, when the Action Phase starts: Equips Pyro Scorpion: Guardian Stance to your Eremite Scorching Loremaster. (When Spirit of Omen: Pyro Scorpion is in play, the character will take 1 less DMG, twice per Round.)",  # noqa: E501
                "zh-CN": "结束阶段：造成1点火元素伤害；在镀金旅团·炽沙叙事人使用过「普通攻击」或「元素战技」的回合中，造成的伤害+1。\n可用次数：2\n\n入场时和行动阶段开始：使我方镀金旅团·炽沙叙事人附属炎之魔蝎·守势。（厄灵·炎之魔蝎在场时每回合2次，使角色受到的伤害-1。）"  # noqa: E501
            }
        }
    },
    "CHARACTOR_STATUS/Pyro Scorpion: Guardian Stance": {
        "names": {
            "en-US": "Pyro Scorpion: Guardian Stance",
            "zh-CN": "炎之魔蝎·守势"
        },
        "descs": {
            "4.3": {
                "en-US": "the character will take 1 less DMG, once per Round.",  # noqa: E501
                "zh-CN": "使角色受到的伤害-1，每回合1次。"
            },
        },
    },
    "CHARACTOR_STATUS/Pyro Scorpion: Guardian Stance_talent": {
        "names": {
            "en-US": "Pyro Scorpion: Guardian Stance",
            "zh-CN": "炎之魔蝎·守势"
        },
        "descs": {
            "4.3": {
                "en-US": "the character will take 1 less DMG, twice per Round.",  # noqa: E501
                "zh-CN": "使角色受到的伤害-1，每回合2次。"
            },
        },
    },
    "SKILL_Eremite Scorching Loremaster_PASSIVE/Spirit of Omen's Power": {
        "names": {
            "en-US": "Spirit of Omen's Power",
            "zh-CN": "厄灵之能"
        },
        "descs": {
            "4.3": {
                "en-US": "(Passive) After this character takes DMG: If this character has no greater than 7 HP, they gain 1 Energy. (Once per Match)",  # noqa: E501
                "zh-CN": "【被动】此角色受到伤害后：如果此角色生命值不多于7，则获得1点充能。（整场牌局限制1次）"  # noqa: E501
            }
        }
    },
    "TALENT_Eremite Scorching Loremaster/Scorpocalypse": {
        "names": {
            "en-US": "Scorpocalypse",
            "zh-CN": "魔蝎烈祸"
        },
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Eremite Scorching Loremaster, equip this card.\nAfter Eremite Scorching Loremaster equips this card, immediately use Spirit of Omen's Awakening: Pyro Scorpion once.\nSpirit of Omen: Pyro Scorpion created by Eremite Scorching Loremaster with this card equipped will deal +1 DMG in Rounds when Eremite Scorching Loremaster has used a Normal Attack or Elemental Skill.\nSpirit of Omen: Pyro Scorpion's DMG reduction effect can now be triggered twice per Round.\n(Your deck must contain Eremite Scorching Loremaster to add this card to it)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为镀金旅团·炽沙叙事人时，装备此牌。\n镀金旅团·炽沙叙事人装备此牌后，立刻使用一次厄灵苏醒·炎之魔蝎。\n装备有此牌的镀金旅团·炽沙叙事人生成的厄灵·炎之魔蝎在镀金旅团·炽沙叙事人使用过「普通攻击」或「元素战技」的回合中，造成的伤害+1；\n厄灵·炎之魔蝎的减伤效果改为每回合至多2次。\n（牌组中包含镀金旅团·炽沙叙事人，才能加入牌组）"  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Talent_Muscleman.png",  # noqa: E501
        "id": 223031
    },
}


register_class(
    PyroScorpionGuardianStance_4_3 | SpiritOfOmenPyroScorpion_4_3
    | Scorpocalypse_4_3 | EremiteScorchingLoremaster_4_3, desc
)



from typing import Any, List, Literal

from ...event import (
    ChangeObjectUsageEventArguments, CreateObjectEventArguments
)

from ....utils.class_registry import register_class

from ...action import Actions, ChangeObjectUsageAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    CreateStatusPassiveSkill, ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Charactor status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Round status, will last for several rounds and disappear
# Skills


class SecretRiteChasmicSoulfarer(ElementalSkillBase):
    name: Literal[
        'Secret Rite: Chasmic Soulfarer'] = 'Secret Rite: Chasmic Soulfarer'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        When talent equipped and level match, increase damage by 1
        """
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        talent = charactor.talent
        if talent is not None:
            status = match.player_tables[self.position.player_idx].charactors[
                self.position.charactor_idx].status
            for s in status:
                if s.name == 'Pactsworn Pathclearer':  # pragma: no branch
                    if s.usage in talent.active_levels:
                        self.damage = 4
                    break
            else:
                raise AssertionError('No Pactsworn Pathclearer status')
        ret = super().get_actions(match)
        self.damage = 3
        return ret


class SacredRiteWolfsSwiftness(ElementalBurstBase):
    name: Literal[
        "Sacred Rite: Wolf's Swiftness"] = "Sacred Rite: Wolf's Swiftness"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.ELECTRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        attack and increase usage
        """
        ret = super().get_actions(match)
        status = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx].status
        for s in status:
            if s.name == 'Pactsworn Pathclearer':  # pragma: no branch
                ret.append(ChangeObjectUsageAction(
                    object_position = s.position,
                    change_usage = 2
                ))
                return ret
        else:
            raise AssertionError('No Pactsworn Pathclearer status')


class LawfulEnforcer(CreateStatusPassiveSkill):
    name: Literal['Lawful Enforcer'] = 'Lawful Enforcer'
    status_name: Literal['Pactsworn Pathclearer'] = 'Pactsworn Pathclearer'


# Talents


class FeatherfallJudgment_3_3(SkillTalent):
    name: Literal['Featherfall Judgment']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Cyno'] = 'Cyno'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
    )
    skill: Literal[
        'Secret Rite: Chasmic Soulfarer'
    ] = 'Secret Rite: Chasmic Soulfarer'
    active_levels: List[int] = [3, 5]


class FeatherfallJudgment_4_2(SkillTalent):
    name: Literal['Featherfall Judgment']
    version: Literal['4.2'] = '4.2'
    charactor_name: Literal['Cyno'] = 'Cyno'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ELECTRO,
        elemental_dice_number = 3,
    )
    skill: Literal[
        'Secret Rite: Chasmic Soulfarer'
    ] = 'Secret Rite: Chasmic Soulfarer'
    active_levels: List[int] = [0, 2, 4, 6]


# charactor base


class Cyno_3_3(CharactorBase):
    name: Literal['Cyno']
    desc: Literal['', 'transform'] = ''
    version: Literal['3.3'] = '3.3'
    element: ElementType = ElementType.ELECTRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | SecretRiteChasmicSoulfarer
        | SacredRiteWolfsSwiftness | LawfulEnforcer
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU
    ]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = "Invoker's Spear",
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            SecretRiteChasmicSoulfarer(),
            SacredRiteWolfsSwiftness(),
            LawfulEnforcer()
        ]

    def _update_desc(self, match: Any):
        status = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx].status
        for s in status:
            if s.name == 'Pactsworn Pathclearer':  # pragma: no branch
                if s.usage >= 4:
                    self.desc = 'transform'
                else:
                    self.desc = ''
                return
        self.desc = ''

    def event_handler_CHANGE_OBJECT_USAGE(
        self, event: ChangeObjectUsageEventArguments, match: Any
    ) -> List[Actions]:
        self._update_desc(match)
        return []

    def event_handler_CREATE_OBJECT(
        self, event: CreateObjectEventArguments, match: Any
    ) -> List[Actions]:
        self._update_desc(match)
        return []


register_class(Cyno_3_3 | FeatherfallJudgment_3_3 | FeatherfallJudgment_4_2)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions

from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Skills


class GuhuaStyle(PhysicalNormalAttackBase):
    name: Literal['Guhua Style'] = 'Guhua Style'
    cost: Cost = PhysicalNormalAttackBase.get_cost(ElementType.HYDRO)


class FatalRainscreen(ElementalSkillBase):
    name: Literal['Fatal Rainscreen'] = 'Fatal Rainscreen'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, application and create object
        """
        ret = [
            self.attack_opposite_active(match, self.damage, self.damage_type),
            self.charge_self(1),
        ]
        ele_app = self.element_application_self(
            match, DamageElementalType.HYDRO
        )
        ret[0].damage_value_list += ele_app.damage_value_list
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        talent = charactor.talent
        if talent is not None:
            ret += [self.create_team_status('Rain Sword', {
                'version': talent.version,
                'usage': 3,
                'max_usage': 3,
            })]
        else:
            ret += [self.create_team_status('Rain Sword', {
                'version': '3.3'
            })]
        return ret


class Raincutter(ElementalBurstBase):
    name: Literal['Raincutter'] = 'Raincutter'
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.HYDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, application and create object
        """
        ret = [
            self.charge_self(-self.cost.charge),
            self.attack_opposite_active(match, self.damage, self.damage_type),
        ]
        ele_app = self.element_application_self(
            match, DamageElementalType.HYDRO
        )
        ret[1].damage_value_list += ele_app.damage_value_list
        return ret + [
            self.create_team_status('Rainbow Bladework')
        ]


# Talents


class TheScentRemained_3_3(SkillTalent):
    name: Literal['The Scent Remained']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal['Xingqiu'] = 'Xingqiu'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 4
    )
    skill: Literal['Fatal Rainscreen'] = 'Fatal Rainscreen'


class TheScentRemained_4_2(TheScentRemained_3_3):
    name: Literal['The Scent Remained']
    version: Literal['4.2'] = '4.2'
    charactor_name: Literal['Xingqiu'] = 'Xingqiu'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.HYDRO,
        elemental_dice_number = 3
    )
    skill: Literal['Fatal Rainscreen'] = 'Fatal Rainscreen'
# charactor base


class Xingqiu_4_1(CharactorBase):
    name: Literal['Xingqiu']
    version: Literal['4.1'] = '4.1'
    element: ElementType = ElementType.HYDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[GuhuaStyle | FatalRainscreen | Raincutter] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            GuhuaStyle(),
            FatalRainscreen(),
            Raincutter(),
        ]


register_class(Xingqiu_4_1 | TheScentRemained_3_3 | TheScentRemained_4_2)



from ...utils import import_all_modules
from .base import StatusBase
from .charactor_status.base import CharactorStatusBase
from .team_status.base import TeamStatusBase


import_all_modules(__file__, __name__)
__all__ = ('StatusBase', 'CharactorStatusBase', 'TeamStatusBase')



from typing import Dict, List, Literal

from .....utils.class_registry import register_class
from ....modifiable_values import DamageValue
from ....event import SkillEndEventArguments
from ....status.charactor_status.base import (
    ElementalInfusionCharactorStatus, RoundCharactorStatus
)
from ....action import (
    Actions, ChangeObjectUsageAction, DrawCardAction, MakeDamageAction, 
    RemoveObjectAction
)
from ....match import Match
from ....struct import Cost
from ....consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    IconType, SkillType, WeaponType
)
from ....charactor.charactor_base import (
    CharactorBase, ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, SkillBase, SkillTalent
)
from .....utils.desc_registry import DescDictType


class ChiselLightMirror_4_3(ElementalInfusionCharactorStatus, 
                            RoundCharactorStatus):
    name: Literal["Chisel-Light Mirror"] = "Chisel-Light Mirror"
    version: Literal['4.3'] = '4.3'
    usage: int = 2
    max_usage: int = 3

    infused_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        """
        When self normal attack, deal 1 dendro damage. If is charged attack,
        duration +1.
        """
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True
        ):
            # not self skill
            return []
        skill: SkillBase = match.get_object(event.action.position)  # type: ignore  # noqa: E501
        assert skill is not None
        if skill.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return []
        target = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        ret: List[MakeDamageAction | ChangeObjectUsageAction] = [
            MakeDamageAction(
                damage_value_list = [
                    DamageValue(
                        position = self.position,
                        damage_type = DamageType.DAMAGE,
                        target_position = target.position,
                        damage = 1,
                        damage_elemental_type = DamageElementalType.DENDRO,
                        cost = Cost(),
                    )
                ]
            ),
        ]
        if (
            match.player_tables[self.position.player_idx].charge_satisfied
            and self.usage < self.max_usage
        ):
            # is charged attack and self usage not full
            ret.append(
                ChangeObjectUsageAction(
                    object_position = self.position,
                    change_usage = 1
                )
            )
        return ret


class UniversalityAnElaborationOnForm(ElementalSkillBase):
    name: Literal[
        "Universality: An Elaboration on Form"
    ] = "Universality: An Elaboration on Form"
    damage: int = 2
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Match) -> List[Actions]:
        return super().get_actions(match) + [
            self.create_charactor_status('Chisel-Light Mirror')
        ]


class ParticularFieldFettersOfPhenomena(ElementalBurstBase):
    name: Literal[
        "Particular Field: Fetters of Phenomena"
    ] = "Particular Field: Fetters of Phenomena"
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Match) -> List[Actions]:
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        status = None
        for s in charactor.status:
            if s.name == 'Chisel-Light Mirror':
                status = s
                break
        if status is None:
            # attack and create 3-stack status
            return super().get_actions(match) + [
                self.create_charactor_status(
                    'Chisel-Light Mirror',
                    { 'usage': 3 }
                )
            ]
        else:
            # attack and change status usage
            has_talent = self.is_talent_equipped(match)
            new_usage = 3 - status.usage
            if has_talent:
                new_usage = 3
            delta_usage = new_usage - status.usage
            self.damage += status.usage
            ret = super().get_actions(match)
            self.damage -= status.usage
            if new_usage > 0:
                # change usage
                ret.append(ChangeObjectUsageAction(
                    object_position = status.position,
                    change_usage = delta_usage
                ))
            if new_usage == 0:
                # remove status
                ret.append(RemoveObjectAction(
                    object_position = status.position
                ))
            if has_talent:
                # draw 1 card
                ret.append(DrawCardAction(
                    player_idx = self.position.player_idx,
                    number = 1,
                    draw_if_filtered_not_enough = True
                ))
            return ret


class Structuration_4_3(SkillTalent):
    name: Literal["Structuration"] = "Structuration"
    version: Literal['4.3'] = '4.3'
    charactor_name: Literal["Alhaitham"] = "Alhaitham"
    skill: Literal[
        "Particular Field: Fetters of Phenomena"
    ] = "Particular Field: Fetters of Phenomena"
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3,
        charge = 2
    )


class Alhaitham_4_3(CharactorBase):
    name: Literal["Alhaitham"]
    version: Literal['4.3'] = '4.3'
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | UniversalityAnElaborationOnForm 
        | ParticularFieldFettersOfPhenomena
    ] = []
    faction: List[FactionType] = [
        FactionType.SUMERU
    ]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Abductive Reasoning',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            UniversalityAnElaborationOnForm(),
            ParticularFieldFettersOfPhenomena(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTOR/Alhaitham": {
        "names": {
            "en-US": "Alhaitham",
            "zh-CN": "艾尔海森"
        },
        "image_path": "cardface/Char_Avatar_Alhatham.png",  # noqa: E501
        "id": 1706,
        "descs": {
            "4.3": {
                "en-US": "",
                "zh-CN": ""
            }
        }
    },
    "SKILL_Alhaitham_NORMAL_ATTACK/Abductive Reasoning": {
        "names": {
            "en-US": "Abductive Reasoning",
            "zh-CN": "溯因反绎法"
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Physical DMG.",
                "zh-CN": "造成2点物理伤害。"
            }
        }
    },
    "SKILL_Alhaitham_ELEMENTAL_SKILL/Universality: An Elaboration on Form": {
        "names": {
            "en-US": "Universality: An Elaboration on Form",
            "zh-CN": "共相·理式摹写"
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 2 Dendro DMG. This character gains Chisel-Light Mirror.",  # noqa: E501
                "zh-CN": "造成2点草元素伤害，本角色附属琢光镜。"
            }
        }
    },
    "CHARACTOR_STATUS/Chisel-Light Mirror": {
        "names": {
            "en-US": "Chisel-Light Mirror",
            "zh-CN": "琢光镜"
        },
        "descs": {
            "4.3": {
                "en-US": "Physical DMG dealt by the character is converted to Dendro DMG.\nAfter this character performs a Normal Attack: Deals 1 Dendro DMG. If this Skill is a Charged Attack, this state's Duration (Rounds) +1.\nDuration (Rounds): 2 (Can stack, max 3 stacks)",  # noqa: E501
                "zh-CN": "角色造成的物理伤害变为草元素伤害。\n角色普通攻击后：造成1点草元素伤害。如果此技能为重击，则使此状态的持续回合+1。\n持续回合：2（可叠加，最多叠加到3回合）"  # noqa: E501
            }
        },
        "image_path": "status/Alhatham_S.png"
    },
    "SKILL_Alhaitham_ELEMENTAL_BURST/Particular Field: Fetters of Phenomena": {
        "names": {
            "en-US": "Particular Field: Fetters of Phenomena",
            "zh-CN": "殊境·显像缚结"
        },
        "descs": {
            "4.3": {
                "en-US": "Deals 4 Dendro DMG, consumes Chisel-Light Mirror, with the DMG bonus based on Chisel-Light Mirror's Duration (Rounds) consumed.\nIf Chisel-Light Mirror Duration (Rounds) consumed is 0/1/2, then apply Chisel-Light Mirror with 3/2/1 Duration (Rounds) to this character.",  # noqa: E501
                "zh-CN": "造成4点草元素伤害；消耗琢光镜，此伤害提升所消耗琢光镜的持续回合值。\n如果消耗琢光镜的持续回合为0/1/2，则为角色附属持续回合为3/2/1的琢光镜。"  # noqa: E501
            }
        }
    },
    "TALENT_Alhaitham/Structuration": {
        "names": {
            "en-US": "Structuration",
            "zh-CN": "正理"
        },
        "descs": {
            "4.3": {
                "en-US": "Combat Action: When your active character is Alhaitham, equip this card.\nAfter Alhaitham equips this card, immediately use Particular Field: Fetters of Phenomena once.\nWhen your Alhaitham, who has this card equipped, uses Particular Field: Fetters of Phenomena, if you expended at least 1 Duration (Rounds) of Chisel-Light Mirror, attach Chisel-Light Mirror with 3 Duration (Rounds) and draw 1 card.\n(You must have Alhaitham in your deck to add this card to your deck.)",  # noqa: E501
                "zh-CN": "战斗行动：我方出战角色为艾尔海森时，装备此牌。\n艾尔海森装备此牌后，立刻使用一次殊境·显像缚结。\n装备有此牌的艾尔海森使用殊境·显像缚结时：如果消耗了持续回合至少为1的琢光镜，则总是附属持续回合为3的琢光镜，并且抓1张牌。\n（牌组中包含艾尔海森，才能加入牌组）"  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Constellation_Alhatham.png",  # noqa: E501
        "id": 217061
    },
}


register_class(
    ChiselLightMirror_4_3 | Structuration_4_3 | Alhaitham_4_3, desc
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import MakeDamageEventArguments, SkillEndEventArguments

from ...summon.base import SwirlChangeSummonBase

from ...action import (
    Actions, CreateObjectAction, SwitchCharactorAction
)
from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType, DieColor, ElementType, ElementalReactionType, 
    FactionType, ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class AutumnWhirlwind_3_8(SwirlChangeSummonBase):
    name: Literal['Autumn Whirlwind'] = 'Autumn Whirlwind'
    version: Literal['3.8'] = '3.8'
    usage: int = 3
    max_usage: int = 3
    damage: int = 1


# Skills


class Chihayaburu(ElementalSkillBase):
    name: Literal['Chihayaburu'] = 'Chihayaburu'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        ret: List[Actions] = [
            # create status first, so it can change element
            self.create_charactor_status('Midare Ranzan: New'),
        ]
        # make damage
        ret.append(self.attack_opposite_active(
            match, self.damage, self.damage_type))
        # switch charactor
        next_charactor = match.player_tables[
            self.position.player_idx].next_charactor_idx()
        if next_charactor is not None:
            ret.append(
                SwitchCharactorAction(
                    player_idx = self.position.player_idx,
                    charactor_idx = next_charactor,
                )
            )
        ret.append(self.charge_self(1))
        return ret


class KazuhaSlash(ElementalBurstBase):
    name: Literal['Kazuha Slash'] = 'Kazuha Slash'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.ANEMO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return [
            self.create_summon('Autumn Whirlwind'),
            self.attack_opposite_active(match, self.damage, self.damage_type),
            self.charge_self(1),
        ]


# Talents


class PoeticsOfFuubutsu_3_8(SkillTalent):
    name: Literal['Poetics of Fuubutsu']
    version: Literal['3.8'] = '3.8'
    charactor_name: Literal['Kaedehara Kazuha'] = 'Kaedehara Kazuha'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.ANEMO,
        elemental_dice_number = 3,
    )
    skill: Literal['Chihayaburu'] = 'Chihayaburu'

    swirl_element = ElementType.NONE

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        when damage made, if it is from kazuha, and has swirl, record element.
        """
        for damage in event.damages:
            if not self.position.check_position_valid(
                damage.final_damage.position, match, 
                player_idx_same = True, charactor_idx_same = True, 
                target_area = ObjectPositionType.SKILL,
            ):
                # not self charactor use skill, do nothing
                continue
            if damage.final_damage.damage_from_element_reaction:
                # damage from elemental reaction, do nothing
                continue
            if damage.elemental_reaction != ElementalReactionType.SWIRL:
                # not swirl, do nothing
                continue
            assert self.swirl_element == ElementType.NONE, (
                'element type already determined'
            )
            # swirl, change element type
            elements = damage.reacted_elements
            assert elements[0] == ElementType.ANEMO, 'First must be anemo'
            self.swirl_element = elements[1]
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        If kazuha did skill and swirled, generate Poetics of Fuubutsu with
        corresponding element.
        """
        element = self.swirl_element
        # regardless of skill situation, reset swirl element
        self.swirl_element = ElementType.NONE
        if element == ElementType.NONE:
            # no swirl, do nothing
            return []
        if not self.position.check_position_valid(
            event.action.position, match, player_idx_same = True,
            charactor_idx_same = True, target_area = ObjectPositionType.SKILL,
            source_area = ObjectPositionType.CHARACTOR,
        ):  # pragma: no cover
            # not same player or charactor, or not skill, or not equipped
            return []
        # this charactor swirl, generate Poetics of Fuubutsu
        return [CreateObjectAction(
            object_name = f'{self.name}: {element.name.capitalize()}',
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectPositionType.TEAM_STATUS,
                id = 0,
            ),
            object_arguments = {}
        )]


# charactor base


class KaedeharaKazuha_3_8(CharactorBase):
    name: Literal['Kaedehara Kazuha']
    version: Literal['3.8'] = '3.8'
    element: ElementType = ElementType.ANEMO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | Chihayaburu | KazuhaSlash
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.SWORD

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Garyuu Bladework',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            Chihayaburu(),
            KazuhaSlash()
        ]


register_class(
    KaedeharaKazuha_3_8 | PoeticsOfFuubutsu_3_8 | AutumnWhirlwind_3_8
)



import json
import os
from src.lpsim import MatchState
from tests.utils_for_test import (
    check_hp, check_usage, get_pidx_cidx, get_test_id_from_command, 
    make_respond, set_16_omni, read_from_log_json
)


def template():
    match, agent_0, agent_1 = read_from_log_json(
        os.path.join(os.path.dirname(__file__), 'jsons', 'test_new_4_transform.json')  # noqa: E501
    )
    # modify hp
    # for i in range(2):
    #     charactors = match.player_tables[i].player_deck_information.charactors  # noqa: E501
    #     for c in charactors:
    #         c.hp = c.max_hp = 30
    # add omnipotent guide
    set_16_omni(match)
    match.start()
    match.step()
    new_commands = [[], []]
    while True:
        if match.need_respond(0):
            agent = agent_0
            nc = new_commands[0]
        elif match.need_respond(1):
            agent = agent_1
            nc = new_commands[1]
        else:
            raise AssertionError('No need respond.')
        # add tests
        enable = [
            # 1,  # hp
            # 2,  # summon
            # 3,  # support
            # 4,  # team status
            # 5,  # charactor status
            # 6,  # hands
            # 7,  # dice
            # 8,  # charge
            # 9,  # summon element
            10,  # charactor desc
        ]
        # 1 for hp
        hp_str = 'TEST 1'
        for i in range(2):
            for c in match.player_tables[i].charactors:
                hp_str += ' ' + str(c.hp)
        if 1 in enable:
            nc.append(hp_str)
        for tnum, table in enumerate(match.player_tables):
            # 2 for summon usage
            summon_str = f'TEST 2 p{tnum}'
            for s in table.summons:
                summon_str += f' {s.usage}'
            if 2 in enable:
                nc.append(summon_str)
            # 3 for support
            support_str = f'TEST 3 p{tnum}'
            for s in table.supports:
                support_str += f' {s.usage}'
            if 3 in enable:
                nc.append(support_str)
            # 4 for team status
            team_str = f'TEST 4 p{tnum}'
            for s in table.team_status:
                team_str += f' {s.usage}'
            if 4 in enable:
                nc.append(team_str)
            for cnum, c in enumerate(table.charactors):
                # 5 for charactor status
                charactor_str = f'TEST 5 p{tnum}c{cnum}'
                for s in c.status:
                    charactor_str += f' {s.usage}'
                if 5 in enable:
                    nc.append(charactor_str)
                cdesc_str = f'TEST 10 p{tnum}c{cnum} |{c.desc}|'
                if c.is_alive:
                    nc.append(cdesc_str)
            # 6 for hands
            hands_str = f'TEST 6 p{tnum} {len(table.hands)}'
            if 6 in enable:
                nc.append(hands_str)
            # 7 for dice
            dice_str = f'TEST 7 p{tnum}'
            for c in table.dice.colors:
                dice_str += f' {c.value}'
            if 7 in enable:
                nc.append(dice_str)
            # 9 for summon element
            summon_element_str = f'TEST 9 p{tnum}'
            for summon in table.summons:
                summon_element_str += ' ' + summon.damage_elemental_type.value
            if 9 in enable:
                nc.append(summon_element_str)
        # 8 for charactor charge
        charge_str = 'TEST 8'
        for i in range(2):
            for c in match.player_tables[i].charactors:
                charge_str += ' ' + str(c.charge)
        if 8 in enable:
            nc.append(charge_str)
        # do tests
        while True:
            nc.append(agent.commands[0])
            cmd = agent.commands[0].strip().split(' ')
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].summons, cmd[3:])
            elif test_id == 3:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].supports, cmd[3:])
            elif test_id == 4:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].team_status, cmd[3:])
            elif test_id == 5:
                pidx, cidx = get_pidx_cidx(cmd)
                check_usage(match.player_tables[pidx].charactors[cidx].status, 
                            cmd[3:])
            elif test_id == 6:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].hands) == int(cmd[3])
            elif test_id == 7:
                pidx = int(cmd[2][1])
                d = {}
                for c in match.player_tables[pidx].dice.colors:
                    d[c.value] = d.get(c.value, 0) + 1
                for c in cmd[3:]:
                    d[c] -= 1
                    if d[c] == 0:
                        del d[c]
                assert len(d) == 0
            elif test_id == 8:
                charges = []
                for table in match.player_tables:
                    for c in table.charactors:
                        charges.append(c.charge)
                assert charges == [int(x) for x in cmd[2:]]
            elif test_id == 9:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].summons) == len(cmd[3:])
                for i, s in enumerate(match.player_tables[pidx].summons):
                    assert s.damage_elemental_type.value == cmd[3 + i]
            elif test_id == 10:
                pidx, cidx = get_pidx_cidx(cmd)
                desc = match.player_tables[pidx].charactors[cidx].desc
                cmd_desc = ' '.join(cmd[3:])[1:-1]
                assert desc == cmd_desc
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR
    print(json.dumps(new_commands, indent = 4))
    open('111.json', 'w').write((json.dumps(new_commands, indent = 4)))


if __name__ == '__main__':
    template()



from ....utils import import_all_modules
from .base import SupportBase
from .companions import CompanionBase
from .items import ItemBase
from .locations import LocationBase


import_all_modules(__file__, __name__)
__all__ = ('SupportBase', 'CompanionBase', 'ItemBase', 'LocationBase')



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...status.team_status.base import ShieldTeamStatus

from ...status.charactor_status.base import ShieldCharactorStatus

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageIncreaseValue

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class StoneStele_3_7(AttackerSummonBase):
    name: Literal['Stone Stele'] = 'Stone Stele'
    version: Literal['3.7'] = '3.7'
    usage: int = 2
    max_usage: int = 2
    damage_elemental_type: DamageElementalType = DamageElementalType.GEO
    damage: int = 1


# Skills


class DominusLapidis(ElementalSkillBase):
    name: Literal['Dominus Lapidis'] = 'Dominus Lapidis'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon('Stone Stele')
        ]


class DominusLapidisStrikingStone(ElementalSkillBase):
    name: Literal[
        'Dominus Lapidis: Striking Stone'] = 'Dominus Lapidis: Striking Stone'
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 5
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon('Stone Stele'),
            self.create_team_status('Jade Shield')
        ]


class PlanetBefall(ElementalBurstBase):
    name: Literal['Planet Befall'] = 'Planet Befall'
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3,
        charge = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack target, and create status for target
        """
        return super().get_actions(match) + [
            self.create_opposite_charactor_status(
                match, 'Petrification', {}
            )
        ]


# Talents


class DominanceOfEarth_3_7(SkillTalent):
    name: Literal['Dominance of Earth']
    version: Literal['3.7'] = '3.7'
    charactor_name: Literal['Zhongli'] = 'Zhongli'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 5
    )
    skill: Literal[
        'Dominus Lapidis: Striking Stone'
    ] = 'Dominus Lapidis: Striking Stone'

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['REAL', 'TEST']
    ) -> DamageIncreaseValue:
        """
        If active charactor has shield, or team status has shield, increase
        geo summon damage
        """
        if (
            self.position.area != ObjectPositionType.CHARACTOR
            or value.position.player_idx != self.position.player_idx
            or value.position.area != ObjectPositionType.SUMMON
            or value.damage_type != DamageType.DAMAGE
            or value.damage_elemental_type != DamageElementalType.GEO
        ):
            # not equipped, or self summon, or not geo damage
            return value
        table = match.player_tables[self.position.player_idx]
        team_status = table.team_status
        have_shield = False
        team_status = table.team_status
        for status in team_status:
            if issubclass(type(status), ShieldTeamStatus):
                have_shield = True
        if not have_shield:
            # no shield team status, check charactor status
            active_charactor = table.get_active_charactor()
            for status in active_charactor.status:
                if issubclass(type(status), ShieldCharactorStatus):
                    have_shield = True
        if not have_shield:
            # no shield
            return value
        # increase damage
        assert mode == 'REAL'
        value.damage += 1
        return value


# charactor base


class Zhongli_3_7(CharactorBase):
    name: Literal['Zhongli']
    version: Literal['3.7'] = '3.7'
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | DominusLapidis 
        | DominusLapidisStrikingStone | PlanetBefall
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.POLEARM

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Rain of Stone',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            DominusLapidis(),
            DominusLapidisStrikingStone(),
            PlanetBefall()
        ]


register_class(Zhongli_3_7 | DominanceOfEarth_3_7 | StoneStele_3_7)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import (
    RoundPrepareEventArguments, SkillEndEventArguments
)

from ...struct import Cost, ObjectPosition

from ...consts import (
    DamageElementalType, DamageType, IconType, ObjectPositionType, SkillType, 
    WeaponType
)

from ...modifiable_values import (
    DamageElementEnhanceValue, DamageIncreaseValue, DamageValue
)

from ...action import MakeDamageAction
from .base import (
    DefendTeamStatus, RoundTeamStatus, ShieldTeamStatus, 
    SwitchActionTeamStatus, UsageTeamStatus, ElementalInfusionTeamStatus
)


class Icicle_3_3(SwitchActionTeamStatus):
    name: Literal['Icicle'] = 'Icicle'
    version: Literal['3.3'] = '3.3'
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def _action(self, match: Any) -> List[MakeDamageAction]:
        """
        attack enemy active charactor
        """
        active_charactor = match.player_tables[
            1 - self.position.player_idx].get_active_charactor()
        return [MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.DAMAGE,
                    target_position = active_charactor.position,
                    damage = 2,
                    damage_elemental_type = DamageElementalType.CRYO,
                    cost = Cost()
                )
            ]
        )]


class IcyQuill_3_7(UsageTeamStatus):
    name: Literal['Icy Quill'] = 'Icy Quill'
    version: Literal['3.7'] = '3.7'
    usage: int = 3
    max_usage: int = 3
    icon_type: Literal[IconType.ATK_UP_ICE] = IconType.ATK_UP_ICE

    talent_usage: int = 0
    talent_max_usage: int = 0

    def renew(self, new_status: 'IcyQuill_3_7') -> None:
        super().renew(new_status)
        self.talent_max_usage = max(
            new_status.talent_max_usage, self.talent_max_usage)
        self.talent_usage = max(
            new_status.talent_usage, self.talent_usage)

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ):
        """
        When in round prepare, reset talent usage
        """
        self.talent_usage = self.talent_max_usage
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL'],
    ) -> DamageIncreaseValue:
        """
        If our charactor skill or summon deal corresponding elemental DMG, 
        increase DMG.
        """
        if self.usage <= 0:
            # no usage, do nothing
            return value
        if value.damage_type != DamageType.DAMAGE:
            # not damage, do nothing
            return value
        if value.position.player_idx != self.position.player_idx:
            # not this player, do nothing
            return value
        if value.damage_elemental_type != DamageElementalType.CRYO:
            # not corresponding elemental DMG, do nothing
            return value
        if value.position.area != ObjectPositionType.SKILL:
            # not charactor, do nothing
            return value
        # increase DMG
        assert mode == 'REAL'
        value.damage += 1
        # decrease usage
        skill = match.get_object(value.position)
        if (
            skill.skill_type == SkillType.NORMAL_ATTACK 
            and self.talent_usage > 0
        ):
            # decrease talent usage first
            self.talent_usage -= 1
        else:
            self.usage -= 1
        return value


class ChonghuasFrostField_3_3(ElementalInfusionTeamStatus, RoundTeamStatus):
    name: Literal["Chonghua's Frost Field"] = "Chonghua's Frost Field"
    desc: Literal['', 'talent'] = ''
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    talent_activated: bool = False

    infused_elemental_type: DamageElementalType = DamageElementalType.CRYO

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.talent_activated:
            self.desc = 'talent'

    def renew(self, object: 'ChonghuasFrostField_3_3'):
        super().renew(object)
        self.talent_activated = object.talent_activated
        if self.talent_activated:
            self.desc = 'talent'

    def charactor_is_using_right_weapon(
        self, match: Any, position: ObjectPosition
    ) -> bool:
        """
        If position charactor using effected three weapon, return True.
        """
        charactor = match.player_tables[position.player_idx].charactors[
            position.charactor_idx]
        return charactor.weapon_type in [
            WeaponType.SWORD, WeaponType.CLAYMORE, WeaponType.POLEARM]

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self, value: DamageElementEnhanceValue, match: Any, 
        mode: Literal['TEST', 'REAL']
    ) -> DamageElementEnhanceValue:
        """
        If weapon type not right, do not enhance.
        """
        if not self.charactor_is_using_right_weapon(match, value.position):
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(
            value, match, mode)

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any,
        mode: Literal['TEST', 'REAL']
    ) -> DamageIncreaseValue:
        """
        When source weapon type is sword, claymore or polearm, and talent
        activated, +1 normal attack DMG.
        """
        if not self.talent_activated:
            # talent not activated, do nothing
            return value
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not corresponding charactor, do nothing
            return value
        if not self.charactor_is_using_right_weapon(match, value.position):
            # not right weapon, do nothing
            return value
        # add damage
        assert mode == 'REAL'
        value.damage += 1
        return value


class IceLotus_3_3(DefendTeamStatus):
    name: Literal['Ice Lotus'] = 'Ice Lotus'
    version: Literal['3.3'] = '3.3'
    usage: int = 2
    max_usage: int = 2
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1


class FortunePreservingTalisman_4_0(UsageTeamStatus):
    name: Literal[
        'Fortune-Preserving Talisman'] = 'Fortune-Preserving Talisman'
    version: Literal['4.0'] = '4.0'
    usage: int = 3
    max_usage: int = 3
    newly_created: bool = True
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def renew(self, new_status: 'FortunePreservingTalisman_4_0') -> None:
        """
        Reset newly created
        """
        super().renew(new_status)
        self.newly_created = new_status.newly_created

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        if self.newly_created:
            # newly created, mark as false and no effect
            self.newly_created = False
            return []
        if event.action.position.player_idx != self.position.player_idx:
            # not our player skill made damage, do nothing
            return []
        assert event.action.position.area == ObjectPositionType.SKILL
        charactor = match.player_tables[self.position.player_idx].charactors[
            event.action.position.charactor_idx]
        if charactor.damage_taken <= 0:
            # no damage, do nothing
            return []
        # heal charactor
        self.usage -= 1
        return [MakeDamageAction(
            damage_value_list = [
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.HEAL,
                    target_position = charactor.position,
                    damage = -2,
                    damage_elemental_type = DamageElementalType.HEAL,
                    cost = Cost(),
                )
            ],
        )]


class CatClawShield_3_3(ShieldTeamStatus):
    name: Literal['Cat-Claw Shield'] = 'Cat-Claw Shield'
    version: Literal['3.3'] = '3.3'
    usage: int = 1
    max_usage: int = 1


class FlowingCicinShield_3_7(ShieldTeamStatus):
    name: Literal['Flowing Cicin Shield'] = 'Flowing Cicin Shield'
    version: Literal['3.7'] = '3.7'
    usage: int = 1
    max_usage: int = 1


register_class(
    Icicle_3_3 | IcyQuill_3_7 | ChonghuasFrostField_3_3 | IceLotus_3_3 
    | FortunePreservingTalisman_4_0 | CatClawShield_3_3 
    | FlowingCicinShield_3_7
)



from src.lpsim.server.consts import DieColor, ObjectPositionType
from src.lpsim.server.deck import Deck
from src.lpsim.server.interaction import (
    SwitchCardRequest, SwitchCardResponse,
    RerollDiceRequest, RerollDiceResponse,
    SwitchCharactorRequest, SwitchCharactorResponse,
    UseCardRequest, UseCardResponse,
    UseSkillRequest, UseSkillResponse,
)
from src.lpsim.server.match import Match
from src.lpsim.server.struct import Cost, ObjectPosition


def test_response_is_valid():
    # mini fake match
    match = Match()
    deck = Deck.from_str('charactor:Nahida')
    match.player_tables[0].charactors.append(deck.charactors[0])
    match.player_tables[1].charactors.append(deck.charactors[0])
    match.player_tables[0].active_charactor_idx = 0
    match.player_tables[1].active_charactor_idx = 0
    req = SwitchCharactorRequest(
        player_idx = 0,
        active_charactor_idx = 0,
        target_charactor_idx = 1,
        dice_colors = [DieColor.CRYO, DieColor.PYRO],
        cost = Cost(
            any_dice_number = 1,
        )
    )
    resp = SwitchCharactorResponse(
        request = req,
        dice_idxs = [0],
    )
    assert resp.is_valid(match)
    resp.dice_idxs = []
    assert not resp.is_valid(match)
    resp.dice_idxs = [-1]
    assert not resp.is_valid(match)
    resp.dice_idxs = [0, 1]
    assert not resp.is_valid(match)
    resp.dice_idxs = [5]
    assert not resp.is_valid(match)
    resp.dice_idxs = []
    resp.request.cost.any_dice_number = 0
    assert resp.is_valid(match)
    req = RerollDiceRequest(
        player_idx = 0,
        colors = [DieColor.CRYO, DieColor.PYRO, DieColor.PYRO, DieColor.PYRO],
        reroll_times = 1,
    )
    resp = RerollDiceResponse(
        request = req,
        reroll_dice_idxs = [0, 1],
    )
    assert resp.is_valid(match)
    resp.reroll_dice_idxs = [0, 0]
    assert not resp.is_valid(match)
    resp.reroll_dice_idxs = [-1, 0]
    assert not resp.is_valid(match)
    resp.reroll_dice_idxs = [0, 4]
    assert not resp.is_valid(match)
    req = SwitchCardRequest(
        player_idx = 0,
        card_names = ['A', 'B', 'A', 'B', 'C'],
        maximum_switch_number = 2,
    )
    resp = SwitchCardResponse(
        request = req,
        card_idxs = [0, 1],
    )
    assert resp.is_valid(match)
    resp.card_idxs = [0, 0]
    assert not resp.is_valid(match)
    resp.card_idxs = [-1, 0]
    assert not resp.is_valid(match)
    resp.card_idxs = [0, 5]
    assert not resp.is_valid(match)
    resp.card_idxs = [0, 1, 2, 3]
    assert not resp.is_valid(match)
    req = UseCardRequest(
        player_idx = 0,
        card_idx = 0,
        dice_colors = [DieColor.CRYO, DieColor.PYRO, DieColor.PYRO, 
                       DieColor.PYRO],
        targets = [],
        cost = Cost(
            same_dice_number = 2,
        )
    )
    resp = UseCardResponse(
        request = req,
        dice_idxs = [1, 2],
        target = None
    )
    assert resp.is_valid(match)
    resp.dice_idxs = [1, 1]
    assert not resp.is_valid(match)
    resp.dice_idxs = [-1, 1]
    assert not resp.is_valid(match)
    resp.dice_idxs = [1, 5]
    assert not resp.is_valid(match)
    resp.dice_idxs = [1, 2, 3]
    assert not resp.is_valid(match)
    resp.dice_idxs = [1, 2]
    target = ObjectPosition(
        player_idx = 0,
        area = ObjectPositionType.HAND,
        id = -1,
    )
    req.targets.append(target)
    assert not resp.is_valid(match)
    req.targets.clear()
    resp.target = target
    assert not resp.is_valid(match)
    req.targets.append(target)
    assert resp.is_valid(match)
    resp.target = resp.target.set_id(321)
    assert not resp.is_valid(match)
    req = UseSkillRequest(
        player_idx = 0,
        charactor_idx = 2,
        skill_idx = 1,
        dice_colors = [DieColor.CRYO, DieColor.PYRO, DieColor.PYRO, 
                       DieColor.PYRO],
        cost = Cost(
            elemental_dice_number = 1,
            elemental_dice_color = DieColor.CRYO,
            any_dice_number = 1,
        )
    )
    resp = UseSkillResponse(
        request = req,
        dice_idxs = [0, 2],
    )
    assert resp.is_valid(match)
    resp.dice_idxs = []
    assert not resp.is_valid(match)
    resp.dice_idxs = [-1, 3]
    assert not resp.is_valid(match)
    resp.dice_idxs = [0, 0]
    assert not resp.is_valid(match)
    resp.dice_idxs = [0, 5]
    assert not resp.is_valid(match)
    resp.dice_idxs = [0, 1, 2]
    assert not resp.is_valid(match)
    resp.dice_idxs = [1, 2]
    assert not resp.is_valid(match)


if __name__ == '__main__':
    test_response_is_valid()



from typing import List, Literal

from ....utils.class_registry import register_class
from ...consts import DieColor
from ...struct import Cost
from .yoimiya_3_8 import Yoimiya_3_8 as Y_3_8
from .yoimiya_3_8 import RyuukinSaxifrage as RS_3_8
from .yoimiya_3_8 import FireworkFlareUp, NiwabiFireDance


class RyuukinSaxifrage(RS_3_8):
    damage: int = 4
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 4,
        charge = 3
    )


class Yoimiya_3_4(Y_3_8):
    version: Literal['3.4']
    skills: List[FireworkFlareUp | NiwabiFireDance | RyuukinSaxifrage] = []

    def _init_skills(self) -> None:
        self.skills = [
            FireworkFlareUp(),
            NiwabiFireDance(),
            RyuukinSaxifrage()
        ]


register_class(Yoimiya_3_4)



import os
from src.lpsim import MatchState
from tests.utils_for_test import (
    check_hp, check_usage, get_pidx_cidx, get_test_id_from_command, 
    make_respond, set_16_omni, read_from_log_json
)


def test_signora():
    match, agent_0, agent_1 = read_from_log_json(
        os.path.join(os.path.dirname(__file__), 'jsons', 'test_signora.json')
    )
    # add omnipotent guide
    set_16_omni(match)
    match.start()
    match.step()
    new_commands = [[], []]

    while True:
        if match.need_respond(0):
            agent = agent_0
            nc = new_commands[0]
        elif match.need_respond(1):
            agent = agent_1
            nc = new_commands[1]
        else:
            raise AssertionError('No need respond.')
        # do tests
        while True:
            nc.append(agent.commands[0])
            cmd = agent.commands[0].strip().split(' ')
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:4], hps[4:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].summons, cmd[3:])
            elif test_id == 3:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].supports, cmd[3:])
            elif test_id == 4:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].team_status, cmd[3:])
            elif test_id == 5:
                pidx, cidx = get_pidx_cidx(cmd)
                check_usage(match.player_tables[pidx].charactors[cidx].status, 
                            cmd[3:])
            elif test_id == 6:
                pidx = int(cmd[2][1])
                assert len(match.player_tables[pidx].hands) == int(cmd[3])
            elif test_id == 7:
                pidx = int(cmd[2][1])
                d = {}
                for c in match.player_tables[pidx].dice.colors:
                    d[c.value] = d.get(c.value, 0) + 1
                for c in cmd[3:]:
                    d[c] -= 1
                    if d[c] == 0:
                        del d[c]
                assert len(d) == 0
            else:
                raise AssertionError(f'Unknown test id {test_id}')
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == '__main__':
    test_signora()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import AttackerSummonBase

from ...modifiable_values import DamageValue
from ...event import MakeDamageEventArguments, RoundEndEventArguments

from ...action import (
    ActionTypes, Actions, ChangeObjectUsageAction, CreateDiceAction, 
    MakeDamageAction
)
from ...struct import Cost

from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE, ELEMENT_TO_DIE_COLOR, DamageElementalType, 
    DamageType, DieColor, ElementType, FactionType, ObjectPositionType, 
    WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    CharactorBase, SkillTalent
)


# Summons


class GossamerSprite_4_2(AttackerSummonBase):
    name: Literal['Gossamer Sprite'] = 'Gossamer Sprite'
    version: Literal['4.2'] = '4.2'
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.DENDRO
    damage: int = 1

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[MakeDamageAction | ChangeObjectUsageAction]:
        ret = super().event_handler_ROUND_END(event, match)
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        make_damage_action = ret[0]
        assert make_damage_action.type == ActionTypes.MAKE_DAMAGE
        make_damage_action.damage_value_list.append(DamageValue(
            position = self.position,
            damage_type = DamageType.HEAL,
            target_position = active_charactor.position,
            damage = -1,
            damage_elemental_type = DamageElementalType.HEAL,
            cost = Cost(),
        ))
        return ret


# Skills


class UniversalDiagnosis(ElementalSkillBase):
    name: Literal['Universal Diagnosis'] = 'Universal Diagnosis'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [
            self.create_summon('Gossamer Sprite')]


class HolisticRevivification(ElementalBurstBase):
    name: Literal['Holistic Revivification'] = 'Holistic Revivification'
    damage: int = 0
    damage_type: DamageElementalType = DamageElementalType.PIERCING
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 4,
        charge = 2
    )

    def get_actions(self, match: Any) -> List[Actions]:
        return [
            self.charge_self(-2),
            self.create_team_status('Pulsing Clarity'),
            self.create_team_status('Seamless Shield'),
        ]


# Talents


class AllThingsAreOfTheEarth_4_2(SkillTalent):
    name: Literal['All Things Are of the Earth']
    version: Literal['4.2'] = '4.2'
    charactor_name: Literal['Baizhu'] = 'Baizhu'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 4,
        charge = 2
    )
    skill: Literal['Holistic Revivification'] = 'Holistic Revivification'

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[CreateDiceAction]:
        """
        When heal made by our shield, create one dice
        """
        source_position = event.action.damage_value_list[0].position
        if (
            self.position.area != ObjectPositionType.CHARACTOR
            or source_position.area != ObjectPositionType.TEAM_STATUS
            or source_position.player_idx != self.position.player_idx
        ):
            # not our team status, or not equipped
            return []
        source = match.get_object(source_position)
        if source.name != 'Seamless Shield':
            # not our shield
            return []
        # create dice
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        return [
            CreateDiceAction(
                player_idx = self.position.player_idx,
                color = ELEMENT_TO_DIE_COLOR[active_charactor.element],
                number = 1,
            )
        ]


# charactor base


class Baizhu_4_2(CharactorBase):
    name: Literal['Baizhu']
    version: Literal['4.2'] = '4.2'
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        ElementalNormalAttackBase | UniversalDiagnosis | HolisticRevivification
    ] = []
    faction: List[FactionType] = [
        FactionType.LIYUE
    ]
    weapon_type: WeaponType = WeaponType.CATALYST

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'The Classics of Acupuncture',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            UniversalDiagnosis(),
            HolisticRevivification(),
        ]


register_class(Baizhu_4_2 | AllThingsAreOfTheEarth_4_2 | GossamerSprite_4_2)



from enum import Enum


class ElementType(str, Enum):
    """
    Enum representing the type of an element.
    """
    NONE = 'NONE'
    CRYO = 'CRYO'
    HYDRO = 'HYDRO'
    PYRO = 'PYRO'
    ELECTRO = 'ELECTRO'
    GEO = 'GEO'
    DENDRO = 'DENDRO'
    ANEMO = 'ANEMO'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


ELEMENT_DEFAULT_ORDER = [
    ElementType.CRYO,
    ElementType.HYDRO,
    ElementType.PYRO,
    ElementType.ELECTRO,
    ElementType.GEO,
    ElementType.DENDRO,
    ElementType.ANEMO,
]


class DieColor(str, Enum):
    """
    Enum representing the color of a die. It can also be called as elemental
    type, elemental attributes. Besides existing element types, there are
    also Onmi, which can be used as any color.

    Attributes:
        CRYO (str): The die color is cryo.
        HYDRO (str): The die color is hydro.
        PYRO (str): The die color is pyro.
        ELECTRO (str): The die color is electro.
        GEO (str): The die color is geo.
        DENDRO (str): The die color is dendro.
        ANEMO (str): The die color is anemo.
        OMNI (str): The die color is omni.
    """
    CRYO = 'CRYO'
    HYDRO = 'HYDRO'
    PYRO = 'PYRO'
    ELECTRO = 'ELECTRO'
    GEO = 'GEO'
    DENDRO = 'DENDRO'
    ANEMO = 'ANEMO'
    OMNI = 'OMNI'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


ELEMENT_TO_DIE_COLOR = {
    ElementType.CRYO: DieColor.CRYO,
    ElementType.HYDRO: DieColor.HYDRO,
    ElementType.PYRO: DieColor.PYRO,
    ElementType.ELECTRO: DieColor.ELECTRO,
    ElementType.GEO: DieColor.GEO,
    ElementType.DENDRO: DieColor.DENDRO,
    ElementType.ANEMO: DieColor.ANEMO,
}
DIE_COLOR_TO_ELEMENT = {
    DieColor.CRYO: ElementType.CRYO,
    DieColor.HYDRO: ElementType.HYDRO,
    DieColor.PYRO: ElementType.PYRO,
    DieColor.ELECTRO: ElementType.ELECTRO,
    DieColor.GEO: ElementType.GEO,
    DieColor.DENDRO: ElementType.DENDRO,
    DieColor.ANEMO: ElementType.ANEMO,
}


class DamageElementalType(str, Enum):
    """
    Enum representing the type of a damage.
    """
    # elemental damage
    CRYO = 'CRYO'
    HYDRO = 'HYDRO'
    PYRO = 'PYRO'
    ELECTRO = 'ELECTRO'
    GEO = 'GEO'
    DENDRO = 'DENDRO'
    ANEMO = 'ANEMO'

    # other damage
    PHYSICAL = 'PHYSICAL'
    PIERCING = 'PIERCING'
    HEAL = 'HEAL'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


ELEMENT_TO_DAMAGE_TYPE = {
    ElementType.NONE: DamageElementalType.PHYSICAL,
    ElementType.CRYO: DamageElementalType.CRYO,
    ElementType.HYDRO: DamageElementalType.HYDRO,
    ElementType.PYRO: DamageElementalType.PYRO,
    ElementType.ELECTRO: DamageElementalType.ELECTRO,
    ElementType.GEO: DamageElementalType.GEO,
    ElementType.DENDRO: DamageElementalType.DENDRO,
    ElementType.ANEMO: DamageElementalType.ANEMO,
}
DAMAGE_TYPE_TO_ELEMENT = {
    DamageElementalType.CRYO: ElementType.CRYO,
    DamageElementalType.HYDRO: ElementType.HYDRO,
    DamageElementalType.PYRO: ElementType.PYRO,
    DamageElementalType.ELECTRO: ElementType.ELECTRO,
    DamageElementalType.GEO: ElementType.GEO,
    DamageElementalType.DENDRO: ElementType.DENDRO,
    DamageElementalType.ANEMO: ElementType.ANEMO,
    DamageElementalType.PHYSICAL: ElementType.NONE,
    DamageElementalType.PIERCING: ElementType.NONE,
    DamageElementalType.HEAL: ElementType.NONE,
}


class DamageType(str, Enum):
    """
    Enum representing the type of a damage.
    """
    DAMAGE = 'DAMAGE'
    HEAL = 'HEAL'
    ELEMENT_APPLICATION = 'ELEMENT_APPLICATION'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class ObjectType(str, Enum):
    """
    Enum representing the type of an object.
    """
    EMPTY = 'EMPTY'
    CHARACTOR = 'CHARACTOR'
    DICE = 'DICE'
    CARD = 'CARD'
    SUMMON = 'SUMMON'
    SUPPORT = 'SUPPORT'
    SKILL = 'SKILL'
    WEAPON = 'WEAPON'
    ARTIFACT = 'ARTIFACT'
    TALENT = 'TALENT'
    CHARACTOR_STATUS = 'CHARACTOR_STATUS'
    TEAM_STATUS = 'TEAM_STATUS'
    ARCANE = 'ARCANE'
    SYSTEM = 'SYSTEM'

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value


class ObjectPositionType(str, Enum):
    """
    Enum representing the position of an object.
    """
    DECK = 'DECK'
    HAND = 'HAND'
    SUMMON = 'SUMMON'
    SUPPORT = 'SUPPORT'
    DICE = 'DICE'
    CHARACTOR = 'CHARACTOR'
    CHARACTOR_STATUS = 'CHARACTOR_STATUS'
    SKILL = 'SKILL'
    TEAM_STATUS = 'TEAM_STATUS'
    SYSTEM = 'SYSTEM'
    INVALID = 'INVALID'

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.value


class FactionType(str, Enum):
    """
    Factions of charactors.
    """

    MONDSTADT = 'MONDSTADT'
    LIYUE = 'LIYUE'
    INAZUMA = 'INAZUMA'
    SUMERU = 'SUMERU'
    FONTAINE = 'FONTAINE'
    NATLAN = 'NATLAN'
    SNEZHNAYA = 'SNEZHNAYA'

    FATUI = 'FATUI'
    MONSTER = 'MONSTER'
    HILICHURL = 'HILICHURL'

    ARKHE_PNEUMA = 'ARKHE_PNEUMA'
    ARKHE_OUSIA = 'ARKHE_OUSIA'

    THE_EREMITES = 'THE_EREMITES'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class WeaponType(str, Enum):
    """
    Types of weapons.
    """

    SWORD = 'SWORD'
    CLAYMORE = 'CLAYMORE'
    POLEARM = 'POLEARM'
    CATALYST = 'CATALYST'
    BOW = 'BOW'
    OTHER = 'OTHER'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class SkillType(str, Enum):
    """
    Types of skills.
    """

    NORMAL_ATTACK = 'NORMAL_ATTACK'
    ELEMENTAL_SKILL = 'ELEMENTAL_SKILL'
    ELEMENTAL_BURST = 'ELEMENTAL_BURST'
    PASSIVE = 'PASSIVE'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class ElementalReactionType(str, Enum):
    """
    Types of elemental reactions.
    """

    NONE = 'NONE'
    MELT = 'MELT'
    VAPORIZE = 'VAPORIZE'
    OVERLOADED = 'OVERLOADED'
    SUPERCONDUCT = 'SUPERCONDUCT'
    ELECTROCHARGED = 'ELECTROCHARGED'
    FROZEN = 'FROZEN'
    SWIRL = 'SWIRL'
    CRYSTALLIZE = 'CRYSTALLIZE'
    BURNING = 'BURNING'
    BLOOM = 'BLOOM'
    QUICKEN = 'QUICKEN'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.value


class CostLabels(int, Enum):
    SWITCH_CHARACTOR = 0x1
    CARD = 0x2
    COMPANION = 0x4
    LOCATION = 0x8
    ITEM = 0x10
    ARTIFACT = 0x20
    WEAPON = 0x40
    TALENT = 0x80
    FOOD = 0x100
    ARCANE = 0x200
    NORMAL_ATTACK = 0x400
    ELEMENTAL_SKILL = 0x800
    ELEMENTAL_BURST = 0x1000
    CHARGED_ATTACK = 0x2000
    PLUNGING_ATTACK = 0x4000
    # equipment is all artifact, all weapon and some talents
    EQUIPMENT = 0x8000
    # event is some action cards, and some talents
    EVENT = 0x10000

    SKILLS = 0x1c00
    SUPPORTS = 0x1c

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name


class IconType(str, Enum):
    """
    Icon types of objects. Used for display. All summons, status, and supports
    should have one of these types.
    """

    # e.g. ???
    BUFF = 'BUFF'
    # e.g. I Haven't lost yet used
    DEBUFF = 'DEBUFF'
    # e.g. Changing Shifts
    SPECIAL = 'SPECIAL'
    # e.g. enemy hydro abyss mage
    DAZE = 'DAZE'
    # e.g. enemy dog
    DOT = 'DOT'
    # Frozen
    FROZEN = 'FROZEN'

    # e.g. Collei talent
    ATK_SELF = 'ATK_SELF'
    # e.g. Tandoori Chicken
    ATK_UP = 'ATK_UP'
    # e.g. kazuha talent
    ATK_UP_FIRE = 'ELEMENT_ATK_UP_FIRE'
    ATK_UP_ELEC = 'ELEMENT_ATK_UP_ELEC'
    ATK_UP_ICE = 'ELEMENT_ATK_UP_ICE'
    ATK_UP_WATER = 'ELEMENT_ATK_UP_WATER'
    ATK_UP_WIND = 'ELEMENT_ATK_UP_WIND'
    ATK_UP_ROCK = 'ELEMENT_ATK_UP_ROCK'
    ATK_UP_GRASS = 'ELEMENT_ATK_UP_GRASS'
    # e.g. keqing, ayaka
    ELEMENT_ENCHANT_FIRE = 'ELEMENT_ENCHANT_FIRE'
    ELEMENT_ENCHANT_ELEC = 'ELEMENT_ENCHANT_ELEC'
    ELEMENT_ENCHANT_ICE = 'ELEMENT_ENCHANT_ICE'
    ELEMENT_ENCHANT_WATER = 'ELEMENT_ENCHANT_WATER'
    ELEMENT_ENCHANT_WIND = 'ELEMENT_ENCHANT_WIND'
    ELEMENT_ENCHANT_ROCK = 'ELEMENT_ENCHANT_ROCK'
    ELEMENT_ENCHANT_GRASS = 'ELEMENT_ENCHANT_GRASS'
    # debuffs
    DEBUFF_ELEMENT_FIRE = 'DEBUFF_ELEMENT_FIRE'
    DEBUFF_ELEMENT_ELEC = 'DEBUFF_ELEMENT_ELEC'
    DEBUFF_ELEMENT_ICE = 'DEBUFF_ELEMENT_ICE'
    DEBUFF_ELEMENT_WATER = 'DEBUFF_ELEMENT_WATER'
    DEBUFF_ELEMENT_WIND = 'DEBUFF_ELEMENT_WIND'
    DEBUFF_ELEMENT_ROCK = 'DEBUFF_ELEMENT_ROCK'
    DEBUFF_ELEMENT_GRASS = 'DEBUFF_ELEMENT_GRASS'
    DEBUFF_ELEMENT_PHYSICS = 'DEBUFF_ELEMENT_PHYSICS'
    # e.g. Lotus Flower Crisp, Ushi
    BARRIER = 'BARRIER'
    # e.g. Crystallize, Unmovable Mountain
    SHIELD = 'SHIELD'
    # e.g. pizza
    HEAL = 'HEAL'
    # e.g. Electro Hypostasis
    REVIVE = 'REVIVE'

    # food card used for charactor, or egg used for team.
    FOOD = 'FOOD'

    # e.g. most of summons
    TIMESTATE = 'TIMESTATE'
    # e.g. Timmie, Liben
    COUNTER = 'COUNTER'

    # e.g. Jade Chamber, Traveler's Handy Sword
    NONE = 'NONE'

    # with others, the status has its special icon based on its name
    OTHERS = 'OTHERS'


ELEMENT_TO_ATK_UP_ICON = {
    ElementType.CRYO: IconType.ATK_UP_ICE,
    ElementType.HYDRO: IconType.ATK_UP_WATER,
    ElementType.PYRO: IconType.ATK_UP_FIRE,
    ElementType.ELECTRO: IconType.ATK_UP_ELEC,
    ElementType.GEO: IconType.ATK_UP_ROCK,
    ElementType.DENDRO: IconType.ATK_UP_GRASS,
    ElementType.ANEMO: IconType.ATK_UP_WIND,
}


ELEMENT_TO_ENCHANT_ICON = {
    ElementType.CRYO: IconType.ELEMENT_ENCHANT_ICE,
    ElementType.HYDRO: IconType.ELEMENT_ENCHANT_WATER,
    ElementType.PYRO: IconType.ELEMENT_ENCHANT_FIRE,
    ElementType.ELECTRO: IconType.ELEMENT_ENCHANT_ELEC,
    ElementType.GEO: IconType.ELEMENT_ENCHANT_ROCK,
    ElementType.DENDRO: IconType.ELEMENT_ENCHANT_GRASS,
    ElementType.ANEMO: IconType.ELEMENT_ENCHANT_WIND,
}


class PlayerActionLabels(int, Enum):
    SKILL = 0x1
    SWITCH = 0x2
    CARD = 0x4
    TUNE = 0x8
    END = 0x10

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.name



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import GameStartEventArguments

from ...action import Actions, ChangeObjectUsageAction, CreateObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DieColor, ElementType, FactionType, 
    ObjectPositionType, WeaponType
)
from ..charactor_base import (
    CreateStatusPassiveSkill, ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Skills


class VolatileSporeCloud(ElementalSkillBase):
    name: Literal['Volatile Spore Cloud'] = 'Volatile Spore Cloud'
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3
    )


class FeatherSpreading(ElementalBurstBase):
    name: Literal['Feather Spreading'] = 'Feather Spreading'
    damage: int = 4
    damage_type: DamageElementalType = DamageElementalType.DENDRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3,
        charge = 2,
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Base on usage, increase damage and reset usage
        """
        charactor = match.player_tables[self.position.player_idx].charactors[
            self.position.charactor_idx]
        found_status: Any = None
        for status in charactor.status:
            if status.name == 'Radical Vitality':  # pragma: no branch
                found_status = status
                break
        else:
            raise AssertionError('Radical Vitality not found')
        if found_status.usage == 0:
            # no usage, do nothing
            return super().get_actions(match)
        # increase damage
        self.damage += found_status.usage
        ret = super().get_actions(match)
        self.damage = 4
        ret.append(ChangeObjectUsageAction(
            object_position = found_status.position,
            change_usage = - found_status.usage
        ))
        return ret


class RadicalVitality(CreateStatusPassiveSkill):
    name: Literal['Radical Vitality'] = 'Radical Vitality'
    status_name: Literal['Radical Vitality'] = 'Radical Vitality'


# Talents


class ProliferatingSpores_3_3(SkillTalent):
    name: Literal['Proliferating Spores']
    version: Literal['3.3'] = '3.3'
    charactor_name: Literal[
        'Jadeplume Terrorshroom'] = 'Jadeplume Terrorshroom'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.DENDRO,
        elemental_dice_number = 3
    )
    skill: Literal['Volatile Spore Cloud'] = 'Volatile Spore Cloud'

    def equip(self, match: Any) -> List[CreateObjectAction]:
        """
        When equip, re-create Radical Vitality with 1 more max usage
        """
        return [
            CreateObjectAction(
                object_name = 'Radical Vitality',
                object_position = self.position.set_area(
                    ObjectPositionType.CHARACTOR_STATUS
                ),
                object_arguments = { 'max_usage': 4 }
            )
        ]


# charactor base


class JadeplumeTerrorshroom_3_3(CharactorBase):
    name: Literal['Jadeplume Terrorshroom']
    version: Literal['3.3'] = '3.3'
    element: ElementType = ElementType.DENDRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | VolatileSporeCloud | FeatherSpreading 
        | RadicalVitality
    ] = []
    faction: List[FactionType] = [
        FactionType.MONSTER
    ]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name = 'Majestic Dance',
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            VolatileSporeCloud(),
            FeatherSpreading(),
            RadicalVitality()
        ]


register_class(JadeplumeTerrorshroom_3_3 | ProliferatingSpores_3_3)



from typing import Dict, List, Literal

from .....utils.desc_registry import DescDictType

from .....utils.class_registry import register_class

from ....charactor.charactor_base import ElementalNormalAttackBase

from ....summon.base import DefendSummonBase

from ....consts import ELEMENT_TO_DAMAGE_TYPE, DamageElementalType

from ....match import Match
from ....charactor.hydro.rhodeia_4_2 import (
    RhodeiaOfLoch_4_2, RhodeiaElementSkill as RES_4_2, TideAndTorrent
)


class Frog_4_3(DefendSummonBase):
    name: Literal['Oceanic Mimic: Frog'] = 'Oceanic Mimic: Frog'
    version: Literal['4.3'] = '4.3'
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.HYDRO
    damage: int = 2
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    attack_until_run_out_of_usage: bool = True


mimic_names = [
    'Oceanic Mimic: Squirrel',
    'Oceanic Mimic: Frog',
    'Oceanic Mimic: Raptor'
]


class RhodeiaElementSkill(RES_4_2):
    version: Literal['4.3'] = '4.3'

    def __init__(self, *argv, **kwargs):
        super().__init__(*argv, **kwargs)
        if self.name == 'Oceanid Mimic Summoning':
            pass
        else:
            assert self.name == 'The Myriad Wilds'
            self.summon_number = 2
            self.cost.elemental_dice_number = 5

    def get_next_summon_names(
        self, match: Match, occupied: List[int]
    ) -> int:
        """
        Get next summon names randomly, but fit the rule that try to avoid
        summoning the same type. And maximum summon number should be 2, if
        exceed, randomly re-generate existing summon.
        """
        summons = match.player_tables[self.position.player_idx].summons
        summon_names = [x.name for x in summons]
        select_idx = []
        exist_idx = []
        for idx, name in enumerate(mimic_names):
            if name not in summon_names and idx not in occupied:
                select_idx.append(idx)
            if name in summon_names:
                exist_idx.append(idx)
        if len(select_idx) <= 1:
            # exist at least 2, get one from not occupied and exist
            select_idx = [x for x in exist_idx if x not in occupied]
        if match.config.recreate_mode:
            sname = match.config.random_object_information['rhodeia'].pop(0)
            s_idx = -1
            for idx, name in enumerate(mimic_names):
                if sname.lower() in name.lower():
                    s_idx = idx
                    break
            else:
                raise AssertionError(f'Unknown summon name {sname}')
            assert s_idx in select_idx
            return s_idx
        return select_idx[int(match._random() * len(select_idx))]


class RhodeiaOfLoch_4_3(RhodeiaOfLoch_4_2):
    version: Literal['4.3'] = '4.3'
    skills: List[
        ElementalNormalAttackBase
        | RhodeiaElementSkill | TideAndTorrent
    ] = []

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = 'Surge',
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            RhodeiaElementSkill(
                name = 'Oceanid Mimic Summoning',
            ),
            RhodeiaElementSkill(
                name = 'The Myriad Wilds',
            ),
            TideAndTorrent(),
        ]


desc: Dict[str, DescDictType] = {
    "CHARACTOR/Rhodeia of Loch": {
        "descs": {
            "4.3": {
                "zh-CN": "",
                "en-US": ""
            }
        },
    },
    "SKILL_Rhodeia of Loch_ELEMENTAL_BURST/Tide and Torrent": {
        "descs": {
            "4.3": {
                "zh-CN": "$SKILL_Rhodeia of Loch_ELEMENTAL_BURST/Tide and Torrent|descs|4.2|zh-CN",  # noqa: E501
                "en-US": "$SKILL_Rhodeia of Loch_ELEMENTAL_BURST/Tide and Torrent|descs|4.2|en-US",  # noqa: E501
            }
        }
    },
    "SKILL_Rhodeia of Loch_ELEMENTAL_SKILL/Oceanid Mimic Summoning": {
        "descs": {
            "4.3": {
                "zh-CN": "随机召唤1种纯水幻形。（最多生成2种，超过时随机生成已存在的纯水幻形）",  # noqa: E501
                "en-US": "Randomly summons 1 Oceanid Mimic (Maximum 2 types of mimics. If exceeded, randomly generate preexisting ones.)."  # noqa: E501
            }
        }
    },
    "SKILL_Rhodeia of Loch_ELEMENTAL_SKILL/The Myriad Wilds": {
        "descs": {
            "4.3": {
                "zh-CN": "随机召唤2种纯水幻形。（最多生成2种，超过时随机生成已存在的纯水幻形）",  # noqa: E501
                "en-US": "Randomly summons 2 Oceanid Mimic (Maximum 2 types of mimics. If exceeded, randomly generate preexisting ones.)."  # noqa: E501
            }
        }
    },
    "SKILL_Rhodeia of Loch_NORMAL_ATTACK/Surge": {
        "descs": {
            "4.3": {
                "zh-CN": "造成1点水元素伤害。",
                "en-US": "Deals 1 Hydro DMG."
            }
        }
    },
    "SUMMON/Oceanic Mimic: Frog": {
        "descs": {
            "4.3": {
                "zh-CN": "我方出战角色受到伤害时：抵消1点伤害。\n可用次数：1，耗尽时不弃置此牌。\n\n结束阶段：如果可用次数已耗尽：弃置此牌，造成2点水元素伤害。",  # noqa: E501
                "en-US": "When your active character takes DMG: Decrease DMG taken by 1. When the Usages are depleted, this card will not be discarded. At the End Phase, if Usage(s) have been depleted: Discard this card, deal 2 Hydro DMG. Usage(s): 1"  # noqa: E501
            }
        },
    },
}


register_class(RhodeiaOfLoch_4_3 | Frog_4_3, desc)



from typing import Dict, Literal

import pytest
from src.lpsim.server.card.support.companions import CompanionBase
from src.lpsim.server.card.event.foods import FoodCardBase
from src.lpsim.server.card.support.base import SupportBase
from src.lpsim.server.summon.base import SummonBase
from src.lpsim.server.charactor.charactor_base import CharactorBase
from src.lpsim.utils.desc_registry import (
    DescDictType, desc_exist, get_desc_patch, update_cost, update_desc
)
from src.lpsim.utils.class_registry import (
    get_class_list_by_base_class, get_instance, register_class
)
from src.lpsim.server.struct import Cost
from src.lpsim.server.object_base import EventCardBase


def test_class_registry():

    class Sleep_1_0(EventCardBase):
        name: Literal['Sleep'] = 'Sleep'
        version: Literal['1.0'] = '1.0'
        cost: Cost = Cost()

    sleep_desc: Dict[str, DescDictType] = {
        'CARD/Sleep': {
            "names": {
                'zh-CN': '睡觉',
                'en-US': 'Sleep',
            }
        }
    }

    with pytest.raises(AssertionError):
        # no desc, cannot register
        register_class(Sleep_1_0)
    with pytest.raises(AssertionError):
        # desc without version info, cannot register
        register_class(Sleep_1_0, sleep_desc)
    sleep_desc['CARD/Sleep']['descs'] = {
        '3.0': {
            'zh-CN': '睡觉',
            'en-US': 'Sleep',
        }
    }
    with pytest.raises(AssertionError):
        # desc with wrong version info, cannot register
        register_class(Sleep_1_0, sleep_desc)
    sleep_desc['CARD/Sleep']['descs'] = {
        '1.0': {
            'zh-CN': '睡觉',
            'en-US': 'Sleep',
        }
    }
    register_class(Sleep_1_0, sleep_desc)
    # duplicate, should print warning
    register_class(Sleep_1_0)
    with pytest.raises(AssertionError):
        # wrong base class, get error
        get_instance(CharactorBase, { 'name': 'Sleep', 'version': '1.0' })
    with pytest.raises(AssertionError):
        # wrong base classes, get error
        get_instance(CharactorBase | SummonBase, 
                     { 'name': 'Sleep', 'version': '1.0' })
    # contain right class, get instance
    ins = get_instance(EventCardBase, { 'name': 'Sleep', 'version': '1.0' })
    assert ins.name == 'Sleep'
    ins = get_instance(SupportBase | EventCardBase, 
                       { 'name': 'Sleep', 'version': '1.0' })
    assert ins.name == 'Sleep'
    ins = get_instance(EventCardBase, { 'name': 'Sleep' })
    assert ins.name == 'Sleep'
    with pytest.raises(AssertionError):
        # wrong version, get error
        get_instance(EventCardBase, { 'name': 'Sleep', 'version': '0.1' })


def test_desc_registry():
    sleep_desc: Dict[str, DescDictType] = {
        'CARD/SleepDesc': {
            "names": {
                'zh-CN': '睡觉',
                'en-US': 'Sleep',
            },
            "id": 999998,
            "image_path": "http://www.baidu.com",
            "descs": {
                '1.0': {
                    'zh-CN': '睡觉',
                    'en-US': 'Sleep',
                }
            }
        }
    }
    assert not desc_exist('CARD', 'SleepDesc', '1.0')
    update_desc(sleep_desc)
    assert desc_exist('CARD', 'SleepDesc', '1.0')
    assert not desc_exist('CARD', 'SleepDesc', '2.0')
    # can add again, but print warning
    update_desc(sleep_desc)
    sleep_desc['CARD/SleepDesc']['descs']['1.0']['zh-CN'] = '睡觉1'
    with pytest.raises(ValueError):
        # duplicate key, raise error
        update_desc(sleep_desc)
    sleep_desc['CARD/SleepDesc']['descs']['1.0']['zh-CN'] = '睡觉'
    sleep_desc['CARD/SleepDesc']['names']['zh-CN'] = '睡觉1'
    with pytest.raises(ValueError):
        # duplicate key, raise error
        update_desc(sleep_desc)
    sleep_patch: Dict[str, DescDictType] = {
        'CARD/SleepDesc': {
            "descs": {
                '2.0': {
                    'zh-CN': '睡觉2',
                    'en-US': 'Sleep2',
                }
            }
        }
    }
    # only update new version descs
    update_desc(sleep_patch)
    assert desc_exist('CARD', 'SleepDesc', '2.0')
    sleep_patch_2: Dict[str, DescDictType] = {
        'CARD/SleepDesc': {
            "descs": {
                '3.0': {
                    'zh-CN': '$CARD/SleepDesc|descs|1.0|zh-CN',
                    'en-US': '$CARD/SleepDesc|descs|1.0|en-US',
                }
            }
        }
    }
    update_desc(sleep_patch_2)
    # use key reference to update
    patch = get_desc_patch()
    assert 'CARD/SleepDesc' in patch
    assert 'descs' in patch['CARD/SleepDesc']
    assert patch['CARD/SleepDesc']['descs']['3.0']['zh-CN'] == '睡觉'
    assert patch['CARD/SleepDesc']['descs']['2.0']['zh-CN'] == '睡觉2'
    assert patch['CARD/SleepDesc']['descs']['1.0']['zh-CN'] == '睡觉'
    assert patch['CARD/SleepDesc']['descs']['3.0']['en-US'] == 'Sleep'
    assert patch['CARD/SleepDesc']['descs']['2.0']['en-US'] == 'Sleep2'
    assert patch['CARD/SleepDesc']['descs']['1.0']['en-US'] == 'Sleep'
    sleep_patch_3: Dict[str, DescDictType] = {
        'CARD/SleepDesc': {
            "descs": {
                '3.0': {
                    'zh-CN': '$CARD/SleepDesc|descs|1.0|zh-CN',
                    'en-US': '$CARD/SleepDesc|descs|1.0|en-US',
                }
            }
        }
    }
    # have duplicate key, print warning
    update_desc(sleep_patch_3)
    sleep_patch_4: Dict[str, DescDictType] = {
        'CARD/SleepDesc': {
            "descs": {
                '5.0': {
                    'zh-CN': '$CARD/SleepDesc|descs|4.0|zh-CN',
                    'en-US': '$CARD/SleepDesc|descs|4.0|en-US',
                }
            }
        }
    }
    with pytest.raises(ValueError):
        # wrong key, raise error
        update_desc(sleep_patch_4)
    empty_name: Dict[str, DescDictType] = {
        'CARD/EmptyName': {
            "names": {
                'zh-CN': '',
            }
        }
    }
    with pytest.raises(ValueError):
        # empty name, raise error
        update_desc(empty_name)
    empty_name_2: Dict[str, DescDictType] = {
        'CARD/EmptyName2': {
            "names": {
                'zh-CN': 'XXX',
            }
        }
    }
    with pytest.raises(ValueError):
        # empty name, raise error
        update_desc(empty_name_2)


def test_get_class_list():
    nlist = get_class_list_by_base_class(SupportBase, '3.3', set(['NRE']))
    assert 'NRE' not in nlist
    assert 'Wangshu Inn' in nlist
    assert len(nlist) == 19  # initial 20 and except NRE
    nlist = get_class_list_by_base_class(SupportBase, '3.3')
    assert 'NRE' in nlist
    assert len(nlist) == 20  # initial 20
    nlist = get_class_list_by_base_class(SupportBase | FoodCardBase, '3.3')
    assert len(nlist) == 28  # initial 20 + 8 food cards
    nlist = get_class_list_by_base_class(FoodCardBase, '4.2')
    assert len(nlist) == 12  # in 4.2, there are 12 food cards
    nlist = get_class_list_by_base_class(FoodCardBase, '3.3', set(['NRE']))
    assert len(nlist) == 8
    nlist = get_class_list_by_base_class(CompanionBase, '3.3')
    assert len(nlist) == 12


def test_register_cost():
    dd_desc: Dict[str, DescDictType] = {
        'CARD/DD': {
            "names": {
                'zh-CN': '滴滴',
                'en-US': 'DD',
            },
            "id": 999998,
            "image_path": "http://www.baidu.com",
            "descs": {
                '1.0': {
                    'zh-CN': '滴滴',
                    'en-US': 'DD',
                }
            }
        }
    }
    dd_cost = Cost(same_dice_number = 1)
    with pytest.raises(ValueError):
        # no desc, cannot register cost
        update_cost('CARD', 'DD', '1.0', dd_cost)
    update_desc(dd_desc)
    p = get_desc_patch()
    assert 'CARD/DD' in p
    assert 'cost' not in p['CARD/DD']
    update_cost('CARD', 'DD', '1.0', dd_cost)
    p = get_desc_patch()
    assert 'CARD/DD' in p
    assert 'cost' in p['CARD/DD']
    assert '1.0' in p['CARD/DD']['cost']
    assert p['CARD/DD']['cost']['1.0']['same_dice_number'] == 1
    assert p['CARD/DD']['cost']['1.0']['any_dice_number'] == 0


if __name__ == "__main__":
    test_class_registry()
    test_desc_registry()
    test_get_class_list()
    test_register_cost()



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...summon.base import DefendSummonBase

from ...modifiable_values import DamageIncreaseValue
from ...event import ReceiveDamageEventArguments, RoundPrepareEventArguments

from ...action import Actions, CreateObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType, DamageType, DieColor, ElementType, FactionType, 
    ObjectPositionType, SkillType, WeaponType
)
from ..charactor_base import (
    ElementalBurstBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, CharactorBase, SkillTalent
)


# Summons


class Ushi_3_6(DefendSummonBase):
    name: Literal['Ushi'] = 'Ushi'
    version: Literal['3.6'] = '3.6'
    usage: int = 1
    max_usage: int = 1
    damage_elemental_type: DamageElementalType = DamageElementalType.GEO
    damage: int = 1
    min_damage_to_trigger: int = 1
    max_in_one_time: int = 1
    attack_until_run_out_of_usage: bool = False

    create_status_triggered: bool = False

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If any of our charactor is attacked and not triggered, Arataki Itto
        gains Superlative Superstrength.
        """
        if self.create_status_triggered:
            # already triggered, no effect
            return []
        if (
            event.final_damage.target_position.player_idx 
            != self.position.player_idx
        ):
            # not attack our player, no effect
            return []
        if event.final_damage.damage_type != DamageType.DAMAGE:
            # not damage, no effect
            return []
        # trigger, gain Superlative Superstrength for first Arataki Itto
        charactors = match.player_tables[self.position.player_idx].charactors
        for charactor in charactors:
            if charactor.is_alive and charactor.name == 'Arataki Itto':
                # found alive itto, gain status
                status_position = charactor.position.set_area(
                    ObjectPositionType.CHARACTOR_STATUS
                )
                self.create_status_triggered = True
                return [CreateObjectAction(
                    object_name = 'Superlative Superstrength',
                    object_position = status_position,
                    object_arguments = {}
                )]
        else:  # pragma: no cover
            # not found alive Itto, no effect
            return []


# Skills


class FightClubLegend(PhysicalNormalAttackBase):
    name: Literal['Fight Club Legend'] = 'Fight Club Legend'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 1,
        any_dice_number = 2,
    )


class MasatsuZetsugiAkaushiBurst(ElementalSkillBase):
    name: Literal[
        'Masatsu Zetsugi: Akaushi Burst!'] = 'Masatsu Zetsugi: Akaushi Burst!'
    damage: int = 1
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack, create Ushi and create status
        """
        ret = super().get_actions(match)
        ret.append(self.create_summon('Ushi'))
        ret += [self.create_charactor_status('Superlative Superstrength')]
        return ret


class RoyalDescentBeholdIttoTheEvil(ElementalBurstBase):
    name: Literal[
        'Royal Descent: Behold, Itto the Evil!'
    ] = 'Royal Descent: Behold, Itto the Evil!'
    damage: int = 5
    damage_type: DamageElementalType = DamageElementalType.GEO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 3,
        charge = 3
    )
    version: Literal['3.6'] = '3.6'

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create status
        """
        return super().get_actions(match) + [
            self.create_charactor_status('Raging Oni King', {
                'version': self.version
            })
        ]


# Talents


class AratakiIchiban_3_6(SkillTalent):
    name: Literal['Arataki Ichiban']
    version: Literal['3.6'] = '3.6'
    charactor_name: Literal['Arataki Itto'] = 'Arataki Itto'
    cost: Cost = Cost(
        elemental_dice_color = DieColor.GEO,
        elemental_dice_number = 1,
        any_dice_number = 2,
    )
    skill: Literal['Fight Club Legend'] = 'Fight Club Legend'


# charactor base


class AratakiItto_3_6(CharactorBase):
    name: Literal['Arataki Itto']
    version: Literal['3.6'] = '3.6'
    element: ElementType = ElementType.GEO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        FightClubLegend | MasatsuZetsugiAkaushiBurst 
        | RoyalDescentBeholdIttoTheEvil
    ] = []
    faction: List[FactionType] = [
        FactionType.INAZUMA
    ]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    normal_attack_this_round: int = 0

    def _init_skills(self) -> None:
        self.skills = [
            FightClubLegend(),
            MasatsuZetsugiAkaushiBurst(),
            RoyalDescentBeholdIttoTheEvil(),
        ]

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Any
    ) -> List[Actions]:
        # reset normal attack counter
        self.normal_attack_this_round = 0
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, 
        match: Any, mode: Literal['TEST', 'REAL']
    ):
        """
        When use normal attack, add counter. 

        For the second time or more using normal attack this round,
        and is charged attack, and has status Superlative Superstrength,
        and equipped talent, add 1 damage.
        """
        if not value.is_corresponding_charactor_use_damage_skill(
            self.position, match, SkillType.NORMAL_ATTACK
        ):
            # not use normal attack
            return value
        # add counter
        self.normal_attack_this_round += 1
        if self.normal_attack_this_round <= 1:
            # first time, no effect
            return value
        if not match.player_tables[self.position.player_idx].charge_satisfied:
            # not charged attack, no effect
            return value
        status_found = False
        for status in self.status:
            if status.name == 'Superlative Superstrength':
                status_found = True
                break
        if not status_found:
            # not have status, no effect
            return value
        if self.talent is None:
            # not equipped talent, no effect
            return value
        # add damage
        value.damage += 1
        return value


register_class(
    AratakiItto_3_6 | AratakiIchiban_3_6 | Ushi_3_6
)



from typing import List, Literal

from ....utils.class_registry import register_class
from ...consts import (
    ElementType, FactionType, WeaponType, DamageElementalType,
)
from ..charactor_base import (
    PhysicalNormalAttackBase, ElementalSkillBase, ElementalBurstBase, 
    CharactorBase
)


class PhysicalMob_1_0(CharactorBase):
    """
    Physical mobs. They cannot carry weapons in default. Their normal 
    attacks are 2 physical DMG, elemental skills are 3 physical DMG, 
    elemental bursts are 2 charges, 3 cost, and 5 physical DMG.
    Their elemeny type is only used to decide dice color when using skills.
    """
    name: Literal['PhysicalMob']
    version: Literal['1.0'] = '1.0'
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | ElementalSkillBase | ElementalBurstBase
    ] = []
    faction: List[FactionType] = []
    weapon_type: WeaponType = WeaponType.OTHER

    def __init__(self, **kwargs):
        super().__init__(**kwargs)  # type: ignore

    def _init_skills(self):
        element_name = self.element.value.lower()
        element_name = element_name[0].upper() + element_name[1:]
        normal_attack = PhysicalNormalAttackBase(
            name = 'Physical Normal Attack',
            damage_type = DamageElementalType.PHYSICAL,
            cost = PhysicalNormalAttackBase.get_cost(self.element),
        )
        elemental_skill = ElementalSkillBase(
            name = 'Physical Skill',
            damage_type = DamageElementalType.PHYSICAL,
            cost = ElementalSkillBase.get_cost(self.element),
        )
        elemental_burst = ElementalBurstBase(
            name = 'Physical Burst',
            damage_type = DamageElementalType.PHYSICAL,
            cost = ElementalBurstBase.get_cost(self.element, 3, 2),
            damage = 5,
        )
        self.skills = [normal_attack, elemental_skill, elemental_burst]


register_class(PhysicalMob_1_0)



"""
Arcane legend cards.
"""

from typing import Any, List, Literal

from ...event import UseCardEventArguments

from ....utils.class_registry import register_class

from ...modifiable_values import DamageValue

from ...action import (
    Actions, ConsumeArcaneLegendAction, CreateDiceAction, CreateObjectAction, 
    DrawCardAction, MakeDamageAction, RemoveCardAction
)

from ...struct import Cost, MultipleObjectPosition, ObjectPosition
from ...consts import (
    ELEMENT_TO_DAMAGE_TYPE, CostLabels, DamageType, ElementType, 
    ObjectPositionType, ObjectType
)
from ...object_base import EventCardBase


class ArcaneLegendBase(EventCardBase):
    type: Literal[ObjectType.ARCANE] = ObjectType.ARCANE
    cost_label: int = (
        CostLabels.CARD.value | CostLabels.ARCANE.value
        | CostLabels.EVENT.value
    )

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction]:
        """
        Consume arcane legend.
        """
        assert target is None
        return [ConsumeArcaneLegendAction(
            player_idx = self.position.player_idx
        )]

    def event_handler_USE_CARD(
        self, event: UseCardEventArguments, match: Any
    ) -> List[Actions]:
        """
        Mainly copied from CardBase with modifications with use_card_success.

        If this card is used, trigger events. But if not use_card_success,
        instead of do nothing, will consume arcane legend.
        """
        actions: List[Actions] = []
        if event.action.card_position.id != self.id:
            # not this card
            return actions
        if self.remove_when_used:  # pragma: no branch
            actions.append(RemoveCardAction(
                position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.HAND,
                    id = self.id,
                ),
                remove_type = 'USED',
            ))
        # target is None, otherwise should match class
        assert (
            event.action.target is None
            or (
                isinstance(event.action.target, MultipleObjectPosition)
                == self.target_position_type_multiple
            )
        )
        if event.use_card_success:
            # use success, add actions of the card
            actions += self.get_actions(
                target = event.action.target,  # type: ignore
                match = match,
            )
        else:
            # use failed, only consume arcane legend
            actions += [ConsumeArcaneLegendAction(
                player_idx = self.position.player_idx
            )]
        return actions


class AncientCourtyard_3_8(ArcaneLegendBase):
    name: Literal['Ancient Courtyard']
    version: Literal['3.8'] = '3.8'
    cost: Cost = Cost(arcane_legend = True)

    def is_valid(self, match: Any) -> bool:
        """
        You must have a character who has already equipped a Weapon or Artifact
        """
        for charactor in match.player_tables[
                self.position.player_idx].charactors:
            if charactor.weapon is not None or charactor.artifact is not None:
                return True
        return False

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        No targets.
        """
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | CreateObjectAction]:
        """
        create team status
        """
        assert target is None
        ret: List[ConsumeArcaneLegendAction | CreateObjectAction] = []
        ret += super().get_actions(target, match)
        ret.append(CreateObjectAction(
            object_name = self.name,
            object_position = ObjectPosition(
                player_idx = self.position.player_idx,
                area = ObjectType.TEAM_STATUS,
                id = 0,
            ),
            object_arguments = {}
        ))
        return ret


class CovenantOfRock_3_8(ArcaneLegendBase):
    name: Literal['Covenant of Rock']
    version: Literal['3.8'] = '3.8'
    cost: Cost = Cost(arcane_legend = True)

    def is_valid(self, match: Any) -> bool:
        """
        Can only be played when you have 0 Elemental Dice left.
        """
        return len(match.player_tables[
            self.position.player_idx].dice.colors) == 0

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        No targets.
        """
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | CreateDiceAction]:
        """
        Generate 2 different Elemental Dice.
        """
        assert target is None
        ret: List[ConsumeArcaneLegendAction | CreateDiceAction] = []
        ret += super().get_actions(target, match)
        ret.append(CreateDiceAction(
            player_idx = self.position.player_idx,
            number = 2,
            different = True
        ))
        return ret


class JoyousCelebration_4_2(ArcaneLegendBase):
    name: Literal['Joyous Celebration']
    version: Literal['4.2'] = '4.2'
    cost: Cost = Cost(
        same_dice_number = 1,
        arcane_legend = True
    )
    apply_no_element_charactor: bool = False

    def is_valid(self, match: Any) -> bool:
        """
        Your active character must be one of the following elemental types to
        play this card: Cryo/Hydro/Pyro/Electro/Dendro
        """
        charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        return charactor.element in [
            ElementType.CRYO, ElementType.HYDRO, ElementType.PYRO,
            ElementType.ELECTRO, ElementType.DENDRO
        ]

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        """
        No targets.
        """
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | MakeDamageAction]:
        """
        apply element to corresponding characters.
        """
        assert target is None
        ret: List[ConsumeArcaneLegendAction | MakeDamageAction] = []
        ret += super().get_actions(target, match)
        damage_action = MakeDamageAction(
            damage_value_list = [],
        )
        active_charactor = match.player_tables[
            self.position.player_idx].get_active_charactor()
        element = active_charactor.element
        for c in match.player_tables[self.position.player_idx].charactors:
            if c.is_defeated:
                continue
            if (
                (not self.apply_no_element_charactor) 
                and len(c.element_application) == 0
            ):
                # not apply to no element charactor
                continue
            damage_action.damage_value_list.append(
                DamageValue(
                    position = self.position,
                    damage_type = DamageType.ELEMENT_APPLICATION,
                    target_position = c.position,
                    damage = 0,
                    damage_elemental_type = ELEMENT_TO_DAMAGE_TYPE[element],
                    cost = self.cost.copy(),
                )
            )
        if len(damage_action.damage_value_list) > 0:
            ret.append(damage_action)
        return ret


class JoyousCelebration_4_0(JoyousCelebration_4_2):
    version: Literal['4.0'] = '4.0'
    apply_no_element_charactor: bool = True


class FreshWindOfFreedom_4_1(ArcaneLegendBase):
    name: Literal['Fresh Wind of Freedom']
    version: Literal['4.1'] = '4.1'
    cost: Cost = Cost(arcane_legend = True)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | CreateObjectAction]:
        """
        Create Wind and Freedom team status.
        """
        assert target is None
        return super().get_actions(target, match) + [
            CreateObjectAction(
                object_name = self.name,
                object_position = ObjectPosition(
                    player_idx = self.position.player_idx,
                    area = ObjectPositionType.TEAM_STATUS,
                    id = -1,
                ),
                object_arguments = {},
            )
        ]


class InEveryHouseAStove_4_2(ArcaneLegendBase):
    name: Literal['In Every House a Stove']
    version: Literal['4.2'] = '4.2'
    cost: Cost = Cost(arcane_legend = True)

    def get_targets(self, match: Any) -> List[ObjectPosition]:
        return []

    def get_actions(
        self, target: ObjectPosition | None, match: Any
    ) -> List[ConsumeArcaneLegendAction | DrawCardAction]:
        """
        Draw cards.
        """
        assert target is None
        return super().get_actions(target, match) + [
            DrawCardAction(
                player_idx = self.position.player_idx,
                number = min(match.round_number, 4),
                draw_if_filtered_not_enough = True
            )
        ]


register_class(
    AncientCourtyard_3_8 | CovenantOfRock_3_8 | JoyousCelebration_4_2
    | FreshWindOfFreedom_4_1 | InEveryHouseAStove_4_2 | JoyousCelebration_4_0
)
